#!/usr/bin/env python
import os
import logging
from collections import OrderedDict

import torch
from torch import nn

logger = logging.getLogger(__name__)


MODEL_PATH = 'your_model_path/clip_visual_encoder'
_MODELS = {
    "ViT-B/16": os.path.join(MODEL_PATH, "vit_b16.pth"),
    "ViT-L/14": os.path.join(MODEL_PATH, "vit_l14.pth"),
    "ViT-L/14_336": os.path.join(MODEL_PATH, "vit_l14_336.pth"),
}


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model, n_head, attn_mask=None):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(OrderedDict([
            ("c_fc", nn.Linear(d_model, d_model * 4)),
            ("gelu", QuickGELU()),
            ("c_proj", nn.Linear(d_model * 4, d_model))
        ]))
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x, return_attn=False):
        self.attn_mask = self.attn_mask.to(dtype=x.dtype, device=x.device) if self.attn_mask is not None else None
        if return_attn:
            return self.attn(x, x, x, need_weights=True, attn_mask=self.attn_mask)
        else:
            return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x, return_attn=False):
        if return_attn:
            x_, attn = self.attention(self.ln_1(x), return_attn=True)
            x = x + x_
            x = x + self.mlp(self.ln_2(x))
            return x, attn
        else:
            x = x + self.attention(self.ln_1(x))
            x = x + self.mlp(self.ln_2(x))
            return x


class Transformer(nn.Module):
    def __init__(
            self, width, layers, heads, 
            clip_return_layer=1, clip_return_interval=1
        ):
        super().__init__()
        self.layers = layers
        self.resblocks = nn.ModuleList()
        for _ in range(layers):
            self.resblocks.append(ResidualAttentionBlock(width, heads))
        self.return_index = []
        for i in range(clip_return_layer):
            self.return_index.append(layers - int(i * clip_return_interval) - 1)
        logger.info(f'Teacher return index: {self.return_index}')

    def forward(self, x):
        attn = None
        z = []
        for idx, blk in enumerate(self.resblocks):
            if idx == self.layers - 1:
                x, attn = blk(x, return_attn=True)
            else:
                x = blk(x)
            if idx in self.return_index:
                z.append(x)
        z = torch.stack(z)
        return z, x, attn


class VisionTransformer(nn.Module):
    def __init__(
        self, input_resolution, patch_size, width, layers, heads, output_dim, 
        clip_return_layer=1, clip_return_interval=1,
    ):
        super().__init__()
        logger.info(f'Return Layer: {clip_return_layer}')
        logger.info(f'Return Interval: {clip_return_interval}')

        self.output_dim = output_dim
        self.conv1 = nn.Conv3d(
            3, width, 
            (1, patch_size, patch_size), 
            (1, patch_size, patch_size), 
            (0, 0, 0), bias=False
        )

        scale = width ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width))
        self.ln_pre = LayerNorm(width)
        
        self.transformer = Transformer(
            width, layers, heads, 
            clip_return_layer=clip_return_layer,
            clip_return_interval=clip_return_interval
        )

        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    def forward(self, x, mask=None):
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        N, C, T, H, W = x.shape
        x = x.permute(0, 2, 3, 4, 1).reshape(N * T, H * W, C)

        x = torch.cat([self.class_embedding.to(x.dtype) + torch.zeros(x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device), x], dim=1)  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        x = self.ln_pre(x)

        if mask is not None:
            cls_tokens = x[:, :1, :]
            x = x[:, 1:]
            x = x.reshape(N, T * H * W, C)
            x = x[~mask].view(N * T, -1, C)
            HW = x.shape[1]
            x = torch.cat([cls_tokens, x], dim=1)
        else:
            HW = H * W

        x = x.permute(1, 0, 2)  # NLD -> LND
        x, x_, attn = self.transformer(x)

        K = x.shape[0]
        x = self.ln_post(x[:, 1:, :, :])  # [HW, NT, C]
        x = x.view(K, HW, N, T, C).permute(0, 2, 3, 1, 4).reshape(K, N, T * HW, C)  # [K, N, THW, C]
        x = x @ self.proj

        # x [K, N, THW, C]
        # attn [NT, HW]
        return x, attn[:, 0, 1:]


def inflate_weight(weight_2d, time_dim, center=True):
    logger.info(f'Init center: {center}')
    if center:
        weight_3d = torch.zeros(*weight_2d.shape)
        weight_3d = weight_3d.unsqueeze(2).repeat(1, 1, time_dim, 1, 1)
        middle_idx = time_dim // 2
        weight_3d[:, :, middle_idx, :, :] = weight_2d
    else:
        weight_3d = weight_2d.unsqueeze(2).repeat(1, 1, time_dim, 1, 1)
        weight_3d = weight_3d / time_dim
    return weight_3d


def load_state_dict(model, state_dict, input_resolution=224, patch_size=16, center=True):
    state_dict_3d = model.state_dict()
    for k in state_dict.keys():
        if k in state_dict_3d.keys() and state_dict[k].shape != state_dict_3d[k].shape:
            if len(state_dict_3d[k].shape) <= 2:
                logger.info(f'Ignore: {k}')
                continue
            logger.info(f'Inflate: {k}, {state_dict[k].shape} => {state_dict_3d[k].shape}')
            time_dim = state_dict_3d[k].shape[2]
            state_dict[k] = inflate_weight(state_dict[k], time_dim, center=center)

    pos_embed_checkpoint = state_dict['positional_embedding']
    embedding_size = pos_embed_checkpoint.shape[-1]
    num_patches = (input_resolution // patch_size) ** 2
    orig_size = int((pos_embed_checkpoint.shape[-2] - 1) ** 0.5)
    new_size = int(num_patches ** 0.5)
    if orig_size != new_size:
        logger.info(f'Pos_emb from {orig_size} to {new_size}')
        extra_tokens = pos_embed_checkpoint[:1]
        pos_tokens = pos_embed_checkpoint[1:]
        pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
        pos_tokens = torch.nn.functional.interpolate(
            pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
        pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(0, 2)
        new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=0)
        state_dict['positional_embedding'] = new_pos_embed
    
    model.load_state_dict(state_dict, strict=False)


def clip_b16(
    input_resolution=224,
    clip_return_layer=6,
    clip_return_interval=1
):
    model = VisionTransformer(
        input_resolution=input_resolution, patch_size=16, 
        width=768, layers=12, heads=12, output_dim=512,
        clip_return_layer=clip_return_layer, 
        clip_return_interval=clip_return_interval
    )
    pretrained = _MODELS["ViT-B/16"]
    logger.info(f"Load pretrained weights from {pretrained}")
    state_dict = torch.load(pretrained, map_location='cpu')
    load_state_dict(model, state_dict, input_resolution=input_resolution, patch_size=16)
    return model.eval()


def clip_l14(
    input_resolution=224,
    clip_return_layer=6,
    clip_return_interval=1
):
    model = VisionTransformer(
        input_resolution=input_resolution, patch_size=14,
        width=1024, layers=24, heads=16, output_dim=768,
        clip_return_layer=clip_return_layer,
        clip_return_interval=clip_return_interval
    )
    pretrained = _MODELS["ViT-L/14"]
    logger.info(f"Load pretrained weights from {pretrained}")
    state_dict = torch.load(pretrained, map_location='cpu')
    load_state_dict(model, state_dict, input_resolution=input_resolution, patch_size=14)
    return model.eval()


def clip_l14_336(
    input_resolution=336,
    clip_return_layer=6,
    clip_return_interval=1
):
    model = VisionTransformer(
        input_resolution=input_resolution, patch_size=14, 
        width=1024, layers=24, heads=16, output_dim=768,
        clip_return_layer=clip_return_layer,
        clip_return_interval=clip_return_interval
    )
    pretrained = _MODELS["ViT-L/14_336"]
    logger.info(f"Load pretrained weights from {pretrained}")
    state_dict = torch.load(pretrained, map_location='cpu')
    load_state_dict(model, state_dict, input_resolution=input_resolution, patch_size=14)
    return model.eval()


def build_clip(config):
    model_cls = config.vision_encoder.clip_teacher
    model = eval(model_cls)(
        input_resolution = config.vision_encoder.clip_img_size,
        clip_return_layer=config.vision_encoder.clip_return_layer,
        clip_return_interval=config.vision_encoder.clip_return_interval,
    )
    return model


if __name__ == '__main__':
    import time
    from fvcore.nn import FlopCountAnalysis
    from fvcore.nn import flop_count_table
    import numpy as np

    seed = 4217
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    num_frames = 4
    
    config = {
        'vision_encoder':
            {
            'clip_teacher': 'clip_b16',
            'clip_img_size': 224,
            'clip_return_layer': 6,
            'clip_return_interval': 1,
        }
    }
    from easydict import EasyDict
    model = build_clip(EasyDict(config))
    # model = clip_b16()
    # print(model)

    # flops = FlopCountAnalysis(model, torch.rand(1, 3, num_frames, 224, 224))
    # s = time.time()
    # print(flop_count_table(flops, max_depth=1))
    # print(time.time()-s)
    output = model(torch.rand(1, 3, num_frames, 224, 224))
    print(output[0].shape, output[1].shape)


import torch
from torch import nn
from transformers import AutoConfig

from .image.configuration_image import LanguageBindImageConfig
from .image.modeling_image import LanguageBindImage
from .image.tokenization_image import LanguageBindImageTokenizer
from .image.processing_image import LanguageBindImageProcessor

from .video.configuration_video import LanguageBindVideoConfig
from .video.modeling_video import LanguageBindVideo
from .video.tokenization_video import LanguageBindVideoTokenizer
from .video.processing_video import LanguageBindVideoProcessor

# from .depth.configuration_depth import LanguageBindDepthConfig
# from .depth.modeling_depth import LanguageBindDepth
# from .depth.tokenization_depth import LanguageBindDepthTokenizer
# from .depth.processing_depth import LanguageBindDepthProcessor

# from .audio.configuration_audio import LanguageBindAudioConfig
# from .audio.modeling_audio import LanguageBindAudio
# from .audio.tokenization_audio import LanguageBindAudioTokenizer
# from .audio.processing_audio import LanguageBindAudioProcessor

# from .thermal.configuration_thermal import LanguageBindThermalConfig
# from .thermal.modeling_thermal import LanguageBindThermal
# from .thermal.tokenization_thermal import LanguageBindThermalTokenizer
# from .thermal.processing_thermal import LanguageBindThermalProcessor



config_dict = {
    # 'thermal': LanguageBindThermalConfig,
    'image': LanguageBindImageConfig,
    'video': LanguageBindVideoConfig,
    # 'depth': LanguageBindDepthConfig,
    # 'audio': LanguageBindAudioConfig
}
model_dict = {
    # 'thermal': LanguageBindThermal,
    'image': LanguageBindImage,
    'video': LanguageBindVideo,
    # 'depth': LanguageBindDepth,
    # 'audio': LanguageBindAudio
}
transform_dict = {
    'video': LanguageBindVideoProcessor,
    # 'audio': LanguageBindAudioProcessor,
    # 'depth': LanguageBindDepthProcessor,
    # 'thermal': LanguageBindThermalProcessor,
    'image': LanguageBindImageProcessor,
}

class LanguageBind(nn.Module):
    def __init__(self, clip_type, use_temp=True, cache_dir='./cache_dir'):
        super(LanguageBind, self).__init__()
        self.use_temp = use_temp
        self.modality_encoder = {}
        self.modality_proj = {}
        self.modality_scale = {}
        self.modality_config = {}
        for k, v in clip_type.items():
            pretrained_ckpt = f'LanguageBind/{v}'
            model = model_dict[k].from_pretrained(pretrained_ckpt, cache_dir=cache_dir)
            self.modality_encoder[k] = model.vision_model
            self.modality_proj[k] = model.visual_projection
            self.modality_scale[k] = model.logit_scale
            self.modality_config[k] = model.config
        self.modality_encoder['language'] = model.text_model
        self.modality_proj['language'] = model.text_projection

        self.modality_encoder = nn.ModuleDict(self.modality_encoder)
        self.modality_proj = nn.ModuleDict(self.modality_proj)

    def forward(self, inputs):
        outputs = {}
        for key, value in inputs.items():
            value = self.modality_encoder[key](**value)[1]
            value = self.modality_proj[key](value)
            value = value / value.norm(p=2, dim=-1, keepdim=True)
            if self.use_temp:
                if key != 'language':
                    value = value * self.modality_scale[key].exp()
            outputs[key] = value
        return outputs

def to_device(x, device):
    out_dict = {k: v.to(device) for k, v in x.items()}
    return out_dict




import copy
import math
from collections import OrderedDict
from typing import Tuple, Union

import clip
import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from timm.models.layers import trunc_normal_
from torch import nn
from torch.utils.checkpoint import checkpoint_sequential


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        # orig_type = x.dtype
        # ret = super().forward(x.type(torch.float32))
        # return ret.type(orig_type)
        return super().forward(x)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(
        self, d_model: int, n_head: int, attn_mask: torch.Tensor = None,
    ):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head,)
        self.ln_1 = LayerNorm(d_model)

        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor):
        self.attn_mask = (
            self.attn_mask.to(dtype=x.dtype, device=x.device)
            if self.attn_mask is not None
            else None
        )
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(
        self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None
    ):
        super().__init__()
        self.width = width
        self.layers = layers
        self.resblocks = nn.Sequential(
            *[ResidualAttentionBlock(width, heads, attn_mask) for _ in range(layers)]
        )

    def forward(self, x: torch.Tensor):
        return self.resblocks(x)


class VisionTransformer(nn.Module):
    def __init__(
        self,
        input_resolution: int,
        patch_size: int,
        width: int,
        layers: int,
        heads: int,
        output_dim: int,
    ):
        super().__init__()
        self.input_resolution = input_resolution
        self.output_dim = output_dim
        self.conv1 = nn.Conv2d(
            in_channels=3,
            out_channels=width,
            kernel_size=patch_size,
            stride=patch_size,
            bias=False,
        )

        scale = width ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(
            scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width)
        )
        self.ln_pre = LayerNorm(width)

        self.transformer = Transformer(width, layers, heads)

        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    def forward(self, x: torch.Tensor):
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat(
            [
                self.class_embedding.to(x.dtype)
                + torch.zeros(
                    x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device
                ),
                x,
            ],
            dim=1,
        )  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD

        x = self.ln_post(x[:, 0, :])

        if self.proj is not None:
            x = x @ self.proj
        return x


class CLIP(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        # vision
        image_resolution: int,
        vision_layers: Union[Tuple[int, int, int, int], int],
        vision_width: int,
        vision_patch_size: int,
        # text
        context_length: int,
        vocab_size: int,
        transformer_width: int,
        transformer_heads: int,
        transformer_layers: int,
    ):
        super().__init__()

        self.context_length = context_length

        # vision_heads = vision_width // 64
        # self.visual = VisionTransformer(
        #     input_resolution=image_resolution,
        #     patch_size=vision_patch_size,
        #     width=vision_width,
        #     layers=vision_layers,
        #     heads=vision_heads,
        #     output_dim=embed_dim
        # )

        # self.transformer = Transformer(
        #     width=transformer_width,
        #     layers=transformer_layers,
        #     heads=transformer_heads,
        #     attn_mask=self.build_attention_mask()
        # )

        # self.vocab_size = vocab_size
        # self.token_embedding = nn.Embedding(vocab_size, transformer_width)
        # self.positional_embedding = nn.Parameter(torch.empty(self.context_length, transformer_width))
        # self.ln_final = LayerNorm(transformer_width)

        # self.text_projection = nn.Parameter(torch.empty(transformer_width, embed_dim))
        # self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

        # self.initialize_parameters()

    def initialize_parameters(self):
        nn.init.normal_(self.token_embedding.weight, std=0.02)
        nn.init.normal_(self.positional_embedding, std=0.01)

        proj_std = (self.transformer.width ** -0.5) * (
            (2 * self.transformer.layers) ** -0.5
        )
        attn_std = self.transformer.width ** -0.5
        fc_std = (2 * self.transformer.width) ** -0.5
        for block in self.transformer.resblocks:
            nn.init.normal_(block.attn.in_proj_weight, std=attn_std)
            nn.init.normal_(block.attn.out_proj.weight, std=proj_std)
            nn.init.normal_(block.mlp.c_fc.weight, std=fc_std)
            nn.init.normal_(block.mlp.c_proj.weight, std=proj_std)

        if self.text_projection is not None:
            nn.init.normal_(self.text_projection, std=self.transformer.width ** -0.5)

    def build_attention_mask(self):
        # lazily create causal attention mask, with full attention between the vision tokens
        # pytorch uses additive attention mask; fill with -inf
        mask = torch.empty(self.context_length, self.context_length)
        mask.fill_(float("-inf"))
        mask.triu_(1)  # zero out the lower diagonal
        return mask

    @property
    def dtype(self):
        return self.visual.conv1.weight.dtype

    def encode_image(self, image):
        return self.visual(image.type(self.dtype))

    def encode_text(self, text):
        x = self.token_embedding(text).type(self.dtype)  # [batch_size, n_ctx, d_model]

        x = x + self.positional_embedding.type(self.dtype)
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x).type(self.dtype)

        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), text.argmax(dim=-1)] @ self.text_projection

        return x

    def forward(self, image, text):
        image_features = self.encode_image(image)
        text_features = self.encode_text(text)

        # normalized features
        image_features = image_features / image_features.norm(dim=1, keepdim=True)
        text_features = text_features / text_features.norm(dim=1, keepdim=True)

        # cosine similarity as logits
        logit_scale = self.logit_scale.exp()
        logits_per_image = logit_scale * image_features @ text_features.t()
        logits_per_text = logits_per_image.t()

        # shape = [global_batch_size, global_batch_size]
        return logits_per_image, logits_per_text


class CrossFramelAttentionBlock(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_head: int,
        attn_mask: torch.Tensor = None,
        droppath=0.0,
        T=0,
    ):
        super().__init__()
        self.T = T

        self.message_fc = nn.Linear(d_model, d_model)
        self.message_ln = LayerNorm(d_model)
        self.message_attn = nn.MultiheadAttention(d_model, n_head,)

        self.attn = nn.MultiheadAttention(d_model, n_head,)
        self.ln_1 = LayerNorm(d_model)

        self.drop_path = DropPath(droppath) if droppath > 0.0 else nn.Identity()
        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor):
        self.attn_mask = (
            self.attn_mask.to(dtype=x.dtype, device=x.device)
            if self.attn_mask is not None
            else None
        )
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x):
        l, bt, d = x.size()
        b = bt // self.T
        x = x.view(l, b, self.T, d)

        msg_token = self.message_fc(x[0, :, :, :])
        msg_token = msg_token.view(b, self.T, 1, d)

        msg_token = msg_token.permute(1, 2, 0, 3).view(self.T, b, d)
        msg_token = msg_token + self.drop_path(
            self.message_attn(
                self.message_ln(msg_token),
                self.message_ln(msg_token),
                self.message_ln(msg_token),
                need_weights=False,
            )[0]
        )
        msg_token = msg_token.view(self.T, 1, b, d).permute(1, 2, 0, 3)

        x = torch.cat([x, msg_token], dim=0)

        x = x.view(l + 1, -1, d)
        x = x + self.drop_path(self.attention(self.ln_1(x)))
        x = x[:l, :, :]
        x = x + self.drop_path(self.mlp(self.ln_2(x)))
        return x


class Transformer(nn.Module):
    def __init__(
        self,
        width: int,
        layers: int,
        heads: int,
        attn_mask: torch.Tensor = None,
        droppath=None,
        use_checkpoint=False,
        T=8,
    ):
        super().__init__()
        self.use_checkpoint = use_checkpoint
        if droppath is None:
            droppath = [0.0 for i in range(layers)]
        self.width = width
        self.layers = layers

        self.resblocks = nn.Sequential(
            *[
                CrossFramelAttentionBlock(width, heads, attn_mask, droppath[i], T)
                for i in range(layers)
            ]
        )

    def forward(self, x: torch.Tensor):
        if not self.use_checkpoint:
            return self.resblocks(x)
        else:
            return checkpoint_sequential(self.resblocks, 3, x)


class CrossFrameCommunicationTransformer(nn.Module):
    def __init__(
        self,
        input_resolution: int,
        patch_size: int,
        width: int,
        layers: int,
        heads: int,
        output_dim: int,
        droppath=None,
        T=8,
        use_checkpoint=False,
    ):
        super().__init__()
        self.input_resolution = input_resolution
        self.output_dim = output_dim

        self.conv1 = nn.Conv2d(
            in_channels=3,
            out_channels=width,
            kernel_size=patch_size,
            stride=patch_size,
            bias=False,
        )

        scale = width ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(
            scale * torch.randn((input_resolution // patch_size) ** 2 + 1, width)
        )
        self.ln_pre = LayerNorm(width)

        ## Attention Blocks
        self.transformer = Transformer(
            width, layers, heads, droppath=droppath, use_checkpoint=use_checkpoint, T=T,
        )
        self.ln_post = LayerNorm(width)
        self.proj = nn.Parameter(scale * torch.randn(width, output_dim))

    def init_weights(self):
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, x: torch.Tensor):
        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat(
            [
                self.class_embedding.to(x.dtype)
                + torch.zeros(
                    x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device
                ),
                x,
            ],
            dim=1,
        )  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)

        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)
        x = self.transformer(x)
        x = x.permute(1, 0, 2)

        cls_x = self.ln_post(x[:, 0, :])

        if self.proj is not None:
            cls_x = cls_x @ self.proj

        return cls_x, x[:, 1:, :]


class MulitHeadAttention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads

        self.scale = qk_scale or head_dim ** -0.5

        self.q_proj = nn.Linear(dim, dim, bias=qkv_bias)
        self.k_proj = nn.Linear(dim, dim, bias=qkv_bias)
        self.v_proj = nn.Linear(dim, dim, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, q, k, v):
        B, N, C = q.shape
        B, M, C = k.shape
        q = (
            self.q_proj(q)
            .reshape(B, N, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )
        k = (
            self.k_proj(k)
            .reshape(B, M, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )
        v = (
            self.v_proj(v)
            .reshape(B, M, self.num_heads, C // self.num_heads)
            .permute(0, 2, 1, 3)
        )

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class PromptGeneratorLayer(nn.Module):
    def __init__(
        self, d_model, nhead, dropout=0.0,
    ):
        super().__init__()
        self.cross_attn = MulitHeadAttention(d_model, nhead, proj_drop=dropout)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)

        self.dropout = nn.Dropout(dropout)

        self.mlp = nn.Sequential(
            nn.Linear(d_model, d_model * 4),
            QuickGELU(),
            nn.Dropout(dropout),
            nn.Linear(d_model * 4, d_model),
        )

    def forward(self, x, visual):
        q = k = v = self.norm1(x)
        x = x + self.cross_attn(q, visual, visual)
        x = x + self.dropout(self.mlp(self.norm3(x)))
        return x


class VideoSpecificPrompt(nn.Module):
    def __init__(
        self, layers=2, embed_dim=512, alpha=0.1,
    ):
        super().__init__()
        self.norm = nn.LayerNorm(embed_dim)
        self.decoder = nn.ModuleList(
            [PromptGeneratorLayer(embed_dim, embed_dim // 64) for _ in range(layers)]
        )
        self.alpha = nn.Parameter(torch.ones(embed_dim) * alpha)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward(self, text, visual):
        B, N, C = visual.shape
        visual = self.norm(visual)
        for layer in self.decoder:
            text = layer(text, visual)


from collections import OrderedDict

from timm.models.layers import trunc_normal_


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = nn.LayerNorm(d_model)
        self.mlp = nn.Sequential(
            OrderedDict(
                [
                    ("c_fc", nn.Linear(d_model, d_model * 4)),
                    ("gelu", QuickGELU()),
                    ("c_proj", nn.Linear(d_model * 4, d_model)),
                ]
            )
        )
        self.ln_2 = nn.LayerNorm(d_model)
        self.attn_mask = attn_mask

    def attention(self, x: torch.Tensor):
        self.attn_mask = (
            self.attn_mask.to(dtype=x.dtype, device=x.device)
            if self.attn_mask is not None
            else None
        )
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class MultiframeIntegrationTransformer(nn.Module):
    def __init__(
        self, T, embed_dim=512, layers=1,
    ):
        super().__init__()
        self.T = T
        transformer_heads = embed_dim // 64
        self.positional_embedding = nn.Parameter(torch.empty(1, T, embed_dim))
        trunc_normal_(self.positional_embedding, std=0.02)
        self.resblocks = nn.Sequential(
            *[
                ResidualAttentionBlock(d_model=embed_dim, n_head=transformer_heads)
                for _ in range(layers)
            ]
        )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, (nn.Linear,)):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.LayerNorm):
            nn.init.zeros_(m.bias)
            nn.init.ones_(m.weight)

    def forward(self, x):
        ori_x = x
        x = x + self.positional_embedding
        x = x.permute(1, 0, 2)
        x = self.resblocks(x)
        x = x.permute(1, 0, 2)
        x = x.type(ori_x.dtype) + ori_x

        return x.mean(dim=1, keepdim=False)


class XCLIP(CLIP):
    def __init__(
        self,
        embed_dim: int,
        # vision
        image_resolution: int,
        vision_layers: Union[Tuple[int, int, int, int], int],
        vision_width: int,
        vision_patch_size: int,
        # text
        context_length: int,
        vocab_size: int,
        transformer_width: int,
        transformer_heads: int,
        transformer_layers: int,
        # video
        T=8,
        droppath=0.0,
        mit_layers=1,
        # prompt
        prompts_alpha=1e-4,
        prompts_layers=1,
        # other
        use_cache=True,
        use_checkpoint=False,
    ):
        super().__init__(
            embed_dim,
            image_resolution,
            vision_layers,
            vision_width,
            vision_patch_size,
            context_length,
            vocab_size,
            transformer_width,
            transformer_heads,
            transformer_layers,
        )

        self.prompts_generator = VideoSpecificPrompt(
            layers=prompts_layers, embed_dim=embed_dim, alpha=prompts_alpha,
        )
        self.use_cache = use_cache
        self.mit = MultiframeIntegrationTransformer(
            T=T, embed_dim=embed_dim, layers=mit_layers,
        )

        dpr = (
            [x.item() for x in torch.linspace(0, droppath, vision_layers)]
            if droppath > 0.0
            else None
        )

        vision_heads = vision_width // 64
        self.visual = CrossFrameCommunicationTransformer(
            input_resolution=image_resolution,
            patch_size=vision_patch_size,
            width=vision_width,
            layers=vision_layers,
            heads=vision_heads,
            output_dim=embed_dim,
            droppath=dpr,
            T=T,
            use_checkpoint=use_checkpoint,
        )

        self.transformer = Transformer(
            width=transformer_width,
            layers=transformer_layers,
            heads=transformer_heads,
            attn_mask=self.build_attention_mask(),
        )
        self.vocab_size = vocab_size
        self.token_embedding = nn.Embedding(vocab_size, transformer_width)
        self.positional_embedding = nn.Parameter(
            torch.empty(self.context_length, transformer_width)
        )
        self.ln_final = LayerNorm(transformer_width)
        self.text_projection = nn.Parameter(torch.empty(transformer_width, embed_dim))
        self.logit_scale = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))

        self.cache_text_features = None
        self.prompts_visual_ln = LayerNorm(vision_width)
        self.prompts_visual_proj = nn.Parameter(torch.randn(vision_width, embed_dim))

        self.initialize_parameters()

    @torch.jit.ignore
    def no_weight_decay_keywords(self):
        return {"positional_embedding"}

    def encode_image(self, image):
        return self.visual(image)

    def encode_text(self, text):
        x = self.token_embedding(text)
        eos_indx = text.argmax(dim=-1)
        K, N1, C = x.shape

        x = x + self.positional_embedding
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x)
        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), eos_indx] @ self.text_projection
        x = x.reshape(K, -1)
        return x

    def encode_video(self, image):
        b, t, c, h, w = image.size()
        image = image.reshape(-1, c, h, w)

        cls_features, img_features = self.encode_image(image)
        img_features = self.prompts_visual_ln(img_features)
        img_features = img_features @ self.prompts_visual_proj

        cls_features = cls_features.view(b, t, -1)
        img_features = img_features.view(b, t, -1, cls_features.shape[-1])

        video_features = self.mit(cls_features)

        return video_features, img_features

    def forward(self, image, **kwargs):
        image = rearrange(image, "b c t h w -> b t c h w")
        video_features, _ = self.encode_video(image)
        return video_features.reshape(*video_features.shape, 1, 1, 1)

    def cache_text(self, text):
        self.eval()
        with torch.no_grad():
            if self.cache_text_features is None:
                self.cache_text_features = self.encode_text(text)
        self.train()
        return self.cache_text_features

    def forward_original(self, image, text):
        b = image.shape[0]
        video_features, img_features = self.encode_video(image)
        img_features = img_features.mean(dim=1, keepdim=False)

        if self.use_cache:
            text_features = self.cache_text(text)
        else:
            text_features = self.encode_text(text)

        text_features = text_features.unsqueeze(0).expand(b, -1, -1)
        text_features = text_features + self.prompts_generator(
            text_features, img_features
        )

        video_features = video_features / video_features.norm(dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)
        logit_scale = self.logit_scale.exp()
        logits = torch.einsum("bd,bkd->bk", video_features, logit_scale * text_features)

        return logits


def build_x_clip_model(
    pretrained_path="./pretrained_weights/k400_32_8.pth",
    droppath=0.0,
    use_checkpoint=False,
    logger=None,
    prompts_alpha=1e-1,
    prompts_layers=2,
    use_cache=True,
    mit_layers=4,
    **kwargs,
):
    state_dict = torch.load(pretrained_path, map_location="cpu")["model"]
    T = int(pretrained_path.split("_")[-1].split(".")[0])
    print(T)
    vit = "visual.proj" in state_dict

    if vit:
        vision_width = state_dict["visual.conv1.weight"].shape[0]
        vision_layers = len(
            [
                k
                for k in state_dict.keys()
                if k.startswith("visual.") and k.endswith(".attn.in_proj_weight")
            ]
        )
        vision_patch_size = state_dict["visual.conv1.weight"].shape[-1]
        grid_size = round(
            (state_dict["visual.positional_embedding"].shape[0] - 1) ** 0.5
        )
        image_resolution = vision_patch_size * grid_size
    else:
        counts: list = [
            len(
                set(
                    k.split(".")[2]
                    for k in state_dict
                    if k.startswith(f"visual.layer{b}")
                )
            )
            for b in [1, 2, 3, 4]
        ]
        vision_layers = tuple(counts)

        vision_width = state_dict["visual.layer1.0.conv1.weight"].shape[0]
        output_width = round(
            (state_dict["visual.attnpool.positional_embedding"].shape[0] - 1) ** 0.5
        )
        vision_patch_size = None
        assert (
            output_width ** 2 + 1
            == state_dict["visual.attnpool.positional_embedding"].shape[0]
        )
        image_resolution = output_width * 32

    embed_dim = state_dict["text_projection"].shape[1]
    context_length = state_dict["positional_embedding"].shape[0]
    vocab_size = state_dict["token_embedding.weight"].shape[0]
    transformer_width = state_dict["ln_final.weight"].shape[0]
    transformer_heads = transformer_width // 64
    transformer_layers = len(
        set(
            k.split(".")[2]
            for k in state_dict
            if k.startswith(f"transformer.resblocks")
        )
    )

    model = XCLIP(
        embed_dim,
        image_resolution,
        vision_layers,
        vision_width,
        vision_patch_size,
        context_length,
        vocab_size,
        transformer_width,
        transformer_heads,
        transformer_layers,
        T=T,
        droppath=droppath,
        mit_layers=mit_layers,
        prompts_alpha=prompts_alpha,
        prompts_layers=prompts_layers,
        use_checkpoint=use_checkpoint,
        use_cache=use_cache,
    )

    for key in ["input_resolution", "context_length", "vocab_size"]:
        if key in state_dict:
            del state_dict[key]

    msg = model.load_state_dict(state_dict, strict=False)

    return model.eval()



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Define conversation format for each training phases and language models.

Modified from LLaVA codebase: https://github.com/haotian-liu/LLaVA/blob/main/llava/conversation.py

NOTE:
- an example of required json format is:
    data = {
        "image": IMAGE_PATH, or "images": LIST of IMAGE_PATH,
        "conversations": [
            {"from": "human", "value": "hello"},
            {"from": "assistant", "value": "Hi, how can I help you today?"},
            {"from": "human", "value": "Who are you?"},
            {"from": "assistant", "value": "I am a multimodal large language model created by FAIR. I can assist you with questions related to images and videos."},
        ]
    }
"""

import copy
from dataclasses import dataclass
from typing import Callable, Dict, List, Union


@dataclass
class Conversation:
    system: str
    conversations: list
    bos_token: str
    sep_system: str
    sep_question: str
    sep_answer: str
    place_image_token: Callable
    image_token: str = "<|image|>"
    pre_system: str = ""
    pre_question: str = ""
    pre_answer: str = ""
    eos_token: str = ""

    # TODO (Maaz): Is there a better name for 'num_patches'. It represents number of vision tokens per image/frame.
    def get_conversation_dict_list(
        self, num_images: int = 1, num_patches: int = 144, media_type: str = "image"
    ) -> List[Dict]:
        """
        Each turn of conversation is a dict with source and target keys.
        """
        conv_dict_list = []
        sys_text = self.pre_system + self.system + self.sep_system
        is_first = True

        if media_type == "multi_image":
            # For multiple interleave images, we keep the <image> tags at the same place as in the original text.
            # However replace <image> in annotations with the self.image_token
            for conversation in self.conversations:
                if conversation["from"] == "human":
                    conversation["value"] = conversation["value"].replace(
                        "<image>", self.image_token * num_patches
                    )
        else:
            # Some annotations already have image tags. remove and add ourself.
            self.conversations[0]["value"] = (
                self.conversations[0]["value"]
                .replace("<image>\n", "")
                .replace("\n<image>", "")
                .replace("<image>", "")
                .replace("<video>\n", "")
                .replace("\n<video>", "")
                .replace("<video>", "")
            )

            self.conversations[0]["value"] = self.place_image_token(
                self.conversations[0]["value"],
                self.image_token,
                num_images * num_patches,
            )
        for conv in self.conversations:
            if is_first and conv["from"] == "assistant":
                continue
            if conv["from"] == "human":
                conv_text = ""
                if is_first:
                    conv_text += sys_text
                conv_text += self.pre_question + conv["value"] + self.sep_question
                conv_dict = {"user": conv_text}
                is_first = False
            elif conv["from"] == "assistant":
                conv_text = self.pre_answer + str(conv["value"]) + self.sep_answer
                conv_dict.update({"assistant": conv_text})
                conv_dict_list.append(conv_dict)
            else:
                raise ValueError(
                    f"conv['from'] must be human or assistant, but got {conv['from']}."
                    "Please fix your jsonl file."
                )
        # Add bos and eos token at the start and end of the coversation (TODO (Maaz): Is there a better way to do it?)
        conv_dict_list[0]["user"] = f"{self.bos_token}{conv_dict_list[0]['user']}"
        conv_dict_list[-1][
            "assistant"
        ] = f"{conv_dict_list[-1]['assistant']}{self.eos_token}"

        return conv_dict_list

    def get_generation_prompt(
        self, prompt: str, num_images: int = 1, num_patches: int = 144
    ):
        if prompt.count("<image>") == num_images:
            prompt = prompt.replace("<image>", self.image_token * num_patches)
        else:
            prompt = (
                prompt.replace("<image>\n", "")
                .replace("\n<image>", "")
                .replace("<image>", "")
                .replace("<video>\n", "")
                .replace("\n<video>", "")
                .replace("<video>", "")
            )
            prompt = self.place_image_token(
                prompt,
                self.image_token,
                num_images * num_patches,
            )

        sys_text = self.bos_token + self.pre_system + self.system + self.sep_system
        return (
            sys_text + self.pre_question + prompt + self.sep_question + self.pre_answer
        )

    def add_conv(self, conv: Union[List, Dict]):
        if isinstance(conv, list):
            self.conversations.extend(conv)
        elif isinstance(conv, dict):
            self.conversations.append(conv)
        else:
            raise ValueError(f"conv must be a list or dict, but got {type(conv)}")

    def copy(self):
        return Conversation(
            system=self.system,
            conversations=copy.deepcopy(self.conversations),
            place_image_token=self.place_image_token,
            bos_token=self.bos_token,
            sep_system=self.sep_system,
            sep_question=self.sep_question,
            sep_answer=self.sep_answer,
            pre_system=self.pre_system,
            pre_question=self.pre_question,
            pre_answer=self.pre_answer,
            image_token=self.image_token,
            eos_token=self.eos_token,
        )


conv_warmup = Conversation(
    system="",
    conversations=[],
    place_image_token=lambda text, image_token, num_image_tokens: image_token
    * num_image_tokens,  # [warmup] ignores question entirely
    bos_token="",
    pre_system="",
    pre_question="",
    pre_answer="",
    sep_system="",
    sep_question="",
    sep_answer="\n",
    eos_token="",
    image_token="<|image|>",
)

conv_plm_sft = Conversation(
    system="You are a helpful language and vision assistant. "
    "You are able to understand the visual content that the user provides, "
    "and assist the user with a variety of tasks using natural language.",
    conversations=[],
    place_image_token=lambda text, image_token, num_image_tokens: (
        image_token * num_image_tokens
    )
    + text,
    bos_token="<|begin_of_text|>",
    pre_system="<|start_header_id|>system<|end_header_id|>\n\n",
    pre_question="<|start_header_id|>user<|end_header_id|>\n\n",
    pre_answer="<|start_header_id|>assistant<|end_header_id|>\n\n",
    sep_system="<|eot_id|>",
    sep_question="<|eot_id|>",
    sep_answer="<|eot_id|>",
    eos_token="<|end_of_text|>",
    image_token="<|image|>",
)


REGISTERED_CONVS = {
    "warmup": conv_warmup,
    "plm_sft": conv_plm_sft,
}



import logging
from ayase.models import Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

class KandinskyMotionModule(PipelineModule):
    name = "kandinsky_motion"
    description = "Video/Camera Motion Analysis using Kandinsky Video Tools (VideoMAE-V2)"
    default_config = {"models_dir": "models"}

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._device = "cpu"
        self._ml_available = False

    def setup(self) -> None:
        try:
            import torch

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info(f"Setting up Kandinsky Motion Predictor on {self._device}...")
            
            # Import vendored module
            try:
                from ayase.third_party.kandinsky.video_motion_predictor.model import VideoMotionPredictor
            except ImportError as e:
                logger.warning(f"Failed to import Kandinsky Video Tools: {e}")
                return

            models_dir = self.config.get("models_dir", "models")

            # Load model from HuggingFace
            # We assume the repo is 'ai-forever/kandinsky-video-motion-predictor' and weights are in 'models/video_motion_predictor'
            try:
                self._model = VideoMotionPredictor.from_pretrained(
                    "ai-forever/kandinsky-video-motion-predictor", 
                    subfolder="models/video_motion_predictor",
                    cache_dir=models_dir
                ).to(self._device).eval()
            except Exception as e_sub:
                logger.info(f"Could not load with subfolder, trying root: {e_sub}")
                self._model = VideoMotionPredictor.from_pretrained(
                    "ai-forever/kandinsky-video-motion-predictor",
                    cache_dir=models_dir
                ).to(self._device).eval()

            self._ml_available = True
            logger.info("Kandinsky Motion Predictor loaded successfully.")

        except Exception as e:
            logger.warning(f"Failed to setup Kandinsky Motion Module: {e}")

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available or not sample.is_video:
            return sample

        try:
            # inference returns a dict: 
            # {
            #   "camera_movement_score": [float], 
            #   "object_movement_score": [float], 
            #   "dynamics_score": [float] 
            # }
            # It expects a list of paths or single path.
            
            results = self._model.inference(str(sample.path))
            
            # Results are lists (batch size 1)
            cam_score = results["camera_movement_score"][0]
            obj_score = results["object_movement_score"][0]
            dyn_score = results["dynamics_score"][0]
            
            # Add to metrics or issues
            if sample.quality_metrics is None:
                from ayase.models import QualityMetrics
                sample.quality_metrics = QualityMetrics()
            
            # Interpret scores (Assuming 0-1 or 0-10 scale?)
            # VideoMAE scores are typically logits or normalized. 
            # The model code had 'targets_norm=[3,3,3]', suggesting outputs might be scaled to 0-3 range?
            # Or maybe just normalized. Let's log them as INFO first.
            
            sample.validation_issues.append(
                ValidationIssue(
                    severity=ValidationSeverity.INFO,
                    message=f"Kandinsky Motion - Dynamics: {dyn_score:.2f}, Camera: {cam_score:.2f}, Object: {obj_score:.2f}",
                    details={
                        "kandinsky_dynamics": dyn_score,
                        "kandinsky_camera": cam_score,
                        "kandinsky_object": obj_score
                    }
                )
            )
            
            # Heuristics based on scores (TBD)
            if dyn_score < 0.1:
                 sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.INFO,
                        message="Video appears static (Low Dynamics Score)",
                        details={"dynamics_score": dyn_score}
                    )
                )

        except Exception as e:
            logger.warning(f"Kandinsky Motion inference failed: {e}")

        return sample



"""Butteraugli module.

Butteraugli is Google's perceptual image distance metric, designed for
evaluating compression quality. It models the human visual system and
is used as the core metric in JPEG XL encoding decisions.

Range: 0+ (lower = better, 0 = identical).
  < 0.5: imperceptible difference
  0.5-1.0: barely noticeable
  1.0-2.0: noticeable
  > 2.0: clearly visible distortion

This is a full-reference metric.

Requires one of:
  - ``jxlpy`` (JPEG XL Python bindings with butteraugli)
  - ``butteraugli`` Python package
"""

import logging
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class ButteraugliModule(ReferenceBasedModule):
    name = "butteraugli"
    description = "Butteraugli perceptual distance (Google/JPEG XL, lower=better)"
    default_config = {
        "subsample": 5,
        "warning_threshold": 2.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 5)
        self.warning_threshold = self.config.get("warning_threshold", 2.0)
        self._ml_available = False
        self._backend = None  # "jxlpy" or "butteraugli"

    def setup(self) -> None:
        # Try jxlpy first (JPEG XL bindings with butteraugli)
        try:
            import jxlpy

            self._backend = "jxlpy"
            self._ml_available = True
            logger.info("Butteraugli module initialised (jxlpy backend)")
            return
        except ImportError:
            pass

        # Try standalone butteraugli package
        try:
            import butteraugli as ba

            self._backend = "butteraugli"
            self._ml_available = True
            logger.info("Butteraugli module initialised (butteraugli backend)")
            return
        except ImportError:
            pass

        # Fallback: OpenCV-based approximation using Laplacian + edge comparison
        self._backend = "approx"
        self._ml_available = True
        logger.info("Butteraugli module initialised (OpenCV approximation fallback)")

    def compute_reference_score(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        try:
            ref_img = cv2.imread(str(reference_path))
            dist_img = cv2.imread(str(sample_path))
            if ref_img is None or dist_img is None:
                return None

            # Align sizes
            h = min(ref_img.shape[0], dist_img.shape[0])
            w = min(ref_img.shape[1], dist_img.shape[1])
            ref_img = cv2.resize(ref_img, (w, h))
            dist_img = cv2.resize(dist_img, (w, h))

            if self._backend == "jxlpy":
                return self._compute_jxlpy(ref_img, dist_img)
            elif self._backend == "butteraugli":
                return self._compute_butteraugli(ref_img, dist_img)
            else:
                return self._compute_approx(ref_img, dist_img)
        except Exception as e:
            logger.debug(f"Butteraugli scoring failed: {e}")
            return None

    def _compute_jxlpy(self, ref_bgr: np.ndarray, dist_bgr: np.ndarray) -> float:
        import jxlpy

        ref_rgb = cv2.cvtColor(ref_bgr, cv2.COLOR_BGR2RGB)
        dist_rgb = cv2.cvtColor(dist_bgr, cv2.COLOR_BGR2RGB)
        return float(jxlpy.butteraugli_distance(ref_rgb, dist_rgb))

    def _compute_butteraugli(self, ref_bgr: np.ndarray, dist_bgr: np.ndarray) -> float:
        import butteraugli as ba

        ref_rgb = cv2.cvtColor(ref_bgr, cv2.COLOR_BGR2RGB)
        dist_rgb = cv2.cvtColor(dist_bgr, cv2.COLOR_BGR2RGB)
        return float(ba.compute(ref_rgb, dist_rgb))

    def _compute_approx(self, ref_bgr: np.ndarray, dist_bgr: np.ndarray) -> float:
        """Approximate butteraugli using frequency-domain perceptual difference.

        This is a rough proxy, not the actual butteraugli algorithm.
        Uses multi-scale edge + colour difference analysis.
        """
        ref_lab = cv2.cvtColor(ref_bgr, cv2.COLOR_BGR2LAB).astype(np.float32)
        dist_lab = cv2.cvtColor(dist_bgr, cv2.COLOR_BGR2LAB).astype(np.float32)

        # Per-channel difference
        diff = np.abs(ref_lab - dist_lab)

        # Weight: L channel matters more than a/b
        weighted = diff[:, :, 0] * 0.6 + diff[:, :, 1] * 0.2 + diff[:, :, 2] * 0.2

        # Edge-weighted (perceptually important near edges)
        ref_gray = cv2.cvtColor(ref_bgr, cv2.COLOR_BGR2GRAY)
        edges = cv2.Canny(ref_gray, 50, 150).astype(np.float32) / 255.0
        edge_weight = 1.0 + edges * 2.0  # 3x weight at edges
        weighted = weighted * edge_weight

        # Map to butteraugli-like scale (rough calibration)
        score = float(np.percentile(weighted, 99)) / 30.0
        return max(0.0, score)

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample
        reference = Path(reference) if not isinstance(reference, Path) else reference
        if not reference.exists():
            return sample

        try:
            if sample.is_video:
                score = self._process_video(sample.path, reference)
            else:
                score = self.compute_reference_score(sample.path, reference)

            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.butteraugli = score

            if score > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"High Butteraugli distance: {score:.2f}",
                        details={"butteraugli": score, "threshold": self.warning_threshold},
                        recommendation="Visible perceptual differences from reference.",
                    )
                )
            logger.debug(f"Butteraugli for {sample.path.name}: {score:.2f}")
        except Exception as e:
            logger.error(f"Butteraugli failed for {sample.path}: {e}")
        return sample

    def _process_video(self, path: Path, ref_path: Path) -> Optional[float]:
        ref_cap = cv2.VideoCapture(str(ref_path))
        dist_cap = cv2.VideoCapture(str(path))
        scores = []
        idx = 0

        while True:
            r1, ref_f = ref_cap.read()
            r2, dist_f = dist_cap.read()
            if not r1 or not r2:
                break
            if idx % self.subsample == 0:
                h = min(ref_f.shape[0], dist_f.shape[0])
                w = min(ref_f.shape[1], dist_f.shape[1])
                ref_r = cv2.resize(ref_f, (w, h))
                dist_r = cv2.resize(dist_f, (w, h))
                try:
                    if self._backend == "jxlpy":
                        s = self._compute_jxlpy(ref_r, dist_r)
                    elif self._backend == "butteraugli":
                        s = self._compute_butteraugli(ref_r, dist_r)
                    else:
                        s = self._compute_approx(ref_r, dist_r)
                    scores.append(s)
                except Exception:
                    pass
            idx += 1

        ref_cap.release()
        dist_cap.release()
        return float(np.mean(scores)) if scores else None



"""MaxVQA — Explainable VQA via Language-Prompted CLIP.

ACM MM 2023 Oral — language-prompted VQA using modified CLIP
for explainable quality scoring with MaxWell dataset.

GitHub: https://github.com/VQAssessment/ExplainableVQA

maxvqa_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class MaxVQAModule(PipelineModule):
    name = "maxvqa"
    description = "MaxVQA explainable language-prompted VQA (ACM MM 2023)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._clip_model = None
        self._clip_processor = None
        self._backend = "heuristic"

    def setup(self) -> None:
        if self.test_mode:
            self._backend = "heuristic"
            return

        try:
            import maxvqa
            self._model = maxvqa
            self._backend = "native"
            logger.info("MaxVQA (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: CLIP-based heuristic
        try:
            import torch
            from transformers import CLIPModel, CLIPProcessor

            device = "cuda" if torch.cuda.is_available() else "cpu"
            self._clip_model = CLIPModel.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir="models"
            ).to(device).eval()
            self._clip_processor = CLIPProcessor.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir="models"
            )
            self._device = device
            self._backend = "clip"
            logger.info(f"MaxVQA (CLIP heuristic) initialised on {device}")
            return
        except (ImportError, Exception):
            pass

        self._backend = "heuristic"
        logger.info("MaxVQA (heuristic) — install maxvqa or transformers for better accuracy")

    def process(self, sample: Sample) -> Sample:
        try:
            if self._backend == "native":
                score = self._process_native(sample)
            elif self._backend == "clip":
                score = self._process_clip(sample)
            else:
                score = self._process_heuristic(sample)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.maxvqa_score = score

        except Exception as e:
            logger.warning(f"MaxVQA failed for {sample.path}: {e}")

        return sample

    def _process_native(self, sample: Sample) -> Optional[float]:
        return float(self._model.predict(str(sample.path)))

    def _process_clip(self, sample: Sample) -> Optional[float]:
        """CLIP-based: cosine similarity with quality anchor texts."""
        import torch
        from PIL import Image

        quality_texts = [
            "a high quality, sharp, well-lit video frame",
            "a low quality, blurry, poorly-lit video frame",
        ]

        frames = self._extract_frames(sample)
        if not frames:
            return None

        scores = []
        with torch.no_grad():
            for frame in frames:
                pil_img = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
                inputs = self._clip_processor(
                    text=quality_texts, images=pil_img,
                    return_tensors="pt", padding=True,
                )
                inputs = {k: v.to(self._device) for k, v in inputs.items()}
                outputs = self._clip_model(**inputs)
                logits = outputs.logits_per_image.softmax(dim=-1)
                # P(high quality)
                scores.append(float(logits[0, 0].cpu()))

        return float(np.mean(scores))

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: multi-attribute quality scoring."""
        frames = self._extract_frames(sample)
        if not frames:
            return None

        scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            sharpness = min(cv2.Laplacian(gray, cv2.CV_64F).var() / 500.0, 1.0)
            contrast = min(gray.std() / 65.0, 1.0)
            brightness = 1.0 - abs(gray.mean() - 127.5) / 127.5

            b, g, r = frame[:, :, 0].astype(float), frame[:, :, 1].astype(float), frame[:, :, 2].astype(float)
            colorfulness = min(
                (np.sqrt((r - g).var() + (0.5 * (r + g) - b).var()) +
                 0.3 * np.sqrt((r - g).mean() ** 2 + (0.5 * (r + g) - b).mean() ** 2)) / 100.0,
                1.0,
            )

            score = 0.35 * sharpness + 0.25 * contrast + 0.20 * brightness + 0.20 * colorfulness
            scores.append(score)

        return float(np.clip(np.mean(scores), 0.0, 1.0))

    def _extract_frames(self, sample: Sample):
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return []
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)
        return frames



from typing import Dict
import random
import re

from torchvision import transforms

from .utils import sample_video

def return_same(x):
    return x

def _bbox_transform_for_padding(bbox, frame):
    w1, h1, w2, h2 = bbox
    width, height = frame.size
    if width == height:
        pass
    elif width > height:
        h1 += (width - height) // 2
        h2 += (width - height) // 2
        height = width
    else:
        w1 += (height - width) // 2
        w2 += (height - width) // 2
        width = height
    new_bbox = [w1 / width, h1 / height, w2 / width, h2 / height]
    new_bbox = [round(i, 2) for i in new_bbox]
    return new_bbox

def _bbox_transform_for_resize(bbox, frame):
    w1, h1, w2, h2 = bbox
    width, height = frame.size
    new_bbox = [w1 / width, h1 / height, w2 / width, h2 / height]
    new_bbox = [round(i, 2) for i in new_bbox]
    return new_bbox

class InAndOutCropAndResize(object):
    """Crop and resize for in_and_out boxes data according to yuchen
    Args:
        size: tuple of (width, height)
    """

    def __init__(self, size):
        self.size = size

    def __call__(self, img):
        """
        Args:
            img (PIL Image): PIL Image
        Returns:
            PIL Image: PIL image.
        """
        w = img.width
        h = img.height
        x0 = int(w * 0.5 - h * 0.375)
        y0 = int(h * 0.125)
        x1 = int(w * 0.5 + h * 0.375)
        y1 = int(h * 0.875)
        img = img.crop((x0, y0, x1, y1)).resize(self.size)
        return img


class ObjectTrackingParser:
    def __init__(
        self,
        n_frames = 8,
        max_objects = 3,
        is_training=True,
    ):
        self.n_frames = n_frames
        self.max_objects = max_objects
        self.is_training = is_training
        self.img_transform = self.get_img_transform()
        # fmt: off
        self.data_temp = {
            "video_file": "/mnt/bn/llmdatalq/jiaxin/hdvila/20230926/saved/saved_video_clips/0076/lOjn__YCec4.624.1104.mp4",
            "frame_indices": [154, 157, 160, 163, 166, 169, 172, 175, 178, 181, 184, 187, 190, 193, 196, 199, 202],
            "objects": {
                "0": {
                    "phrase": "person",
                    "all_frame_bounding_boxes": [[2, 0, 255, 250], [17, 0, 255, 251], [35, 0, 255, 253], [44, 0, 255, 255], [52, 0, 255, 255], [54, 0, 255, 255], [63, 0, 255, 255], [60, 0, 255, 255], [54, 0, 253, 255], [43, 0, 250, 255], [36, 1, 249, 255], [36, 0, 252, 254], [41, 0, 252, 254], [61, 0, 255, 253], [68, 4, 255, 255], [74, 8, 255, 255], [91, 3, 255, 255]]
                }
            },
            "task": "object_tracking",
            "dataset": "hdvila"
        }
        # fmt: on

    def check_format(self, data_dict: Dict, image_processing_config: Dict):
        # box tracking 数据不支持 do_crop！！！
        if image_processing_config.get('do_crop', False):
            raise ValueError(f'do_crop is not supported in ObjectTrackingParser!')

    def transform(self, data_dict: Dict, image_processing_config: Dict = None) -> Dict:
        self.check_format(data_dict, image_processing_config)

        bbox_transform = _bbox_transform_for_padding if image_processing_config['do_padding'] else _bbox_transform_for_resize

        # sample n_frames
        if isinstance(self.n_frames, int):
            n_frames = self.n_frames
        else:
            n_frames = random.choice(self.n_frames)
        total_frames = list(range(len(data_dict['frame_indices'])))
        idxs = random.sample(total_frames, min(n_frames, len(total_frames)))
        idxs.sort()

        frame_indices = [data_dict['frame_indices'][i] for i in idxs]
        frames = sample_video(data_dict['video_file'], frame_indices=frame_indices)
        img_transform = self.img_transform[data_dict['dataset']]
        frames = [img_transform(f) for f in frames]

        objects = []
        for _, o in data_dict['objects'].items():
            if o is None:
                continue
            all_frame_bounding_boxes = [o['all_frame_bounding_boxes'][i] for i in idxs]
            all_frame_bounding_boxes_t = []
            for bbox, frame in zip(all_frame_bounding_boxes, frames):
                all_frame_bounding_boxes_t.append(bbox_transform(bbox, frame))
            objects.append(all_frame_bounding_boxes_t)
            if len(objects) >= self.max_objects:
                break

        prompt = "Given the bounding box coordinates of these objects in the first frame, output the bounding box coordinates in the following frames.\n{}"
        response = ''

        object_info = ''
        for i, o in enumerate(objects):
            object_info += f'object {i+1}: {o[0]}\n'
            response += f'object {i+1}: {o[1:]}\n'
        response = response.strip()
        prompt = prompt.format(object_info)

        messages = [
            {
                "role": "user",
                "content": [
                    {"type": "video", "video": frames},
                    {"type": "text", "text": prompt}
                ]
            },
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": response}
                ]
            }
        ]
        
        return messages

    def get_img_transform(self):
        return {
            'webvid': return_same,
            'hdvila': transforms.Compose([
                transforms.Resize(size=256),
                transforms.CenterCrop(size=(256, 256))
            ]),
            'hdvila_in_and_out_boxes': InAndOutCropAndResize(size=(256, 256))
        }



# Copyright (c) Meta Platforms, Inc. and affiliates.

import logging
import os
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, cast

import tiktoken
import torch
from tiktoken.load import load_tiktoken_bpe

from ...core.data.conversation import REGISTERED_CONVS
from ...core.tokenizer import TikTokenTokenizer, Tokenizer

logger = logging.getLogger(__name__)


class Llama3Tokenizer(TikTokenTokenizer):
    """
    Tokenizing and encoding/decoding text using the Tiktoken tokenizer.
    """

    special_tokens: Dict[str, int]

    num_reserved_special_tokens = 256

    pat_str = r"(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"  # noqa: E501

    def __init__(self, model_path: str):
        """
        Initializes the Tokenizer with a Tiktoken model.

        Args:
            model_path (str): The path to the Tiktoken model file.
        """
        assert os.path.isfile(model_path), model_path

        mergeable_ranks = load_tiktoken_bpe(model_path)
        num_base_tokens = len(mergeable_ranks)
        special_tokens = [
            "<|begin_of_text|>",
            "<|end_of_text|>",
            "<|image|>",
            "<|reserved_special_token_1|>",
            "<|reserved_special_token_2|>",
            "<|reserved_special_token_3|>",
            "<|start_header_id|>",
            "<|end_header_id|>",
            "<|reserved_special_token_4|>",
            "<|eot_id|>",  # End of turn
        ] + [
            f"<|reserved_special_token_{i}|>"
            for i in range(5, self.num_reserved_special_tokens - 5)
        ]
        self.special_tokens = {
            token: num_base_tokens + i for i, token in enumerate(special_tokens)
        }
        self.model = tiktoken.Encoding(
            name=Path(model_path).name,
            pat_str=self.pat_str,
            mergeable_ranks=mergeable_ranks,
            special_tokens=self.special_tokens,
        )
        logger.info(f"Reloaded tiktoken model from {model_path}")

        self.n_words: int = self.model.n_vocab
        # BOS / EOS token IDs
        self.bos_id: int = self.special_tokens["<|begin_of_text|>"]
        self.eos_id: int = self.special_tokens["<|end_of_text|>"]
        self.eot_id: int = self.special_tokens["<|eot_id|>"]
        self.pad_id: int = -1
        self.stop_tokens = {
            self.special_tokens["<|end_of_text|>"],
            self.special_tokens["<|eot_id|>"],
        }
        logger.info(
            f"#words: {self.n_words} - BOS ID: {self.bos_id} - EOS ID: {self.eos_id}"
        )

    def encode(
        self,
        s: str,
        add_bos: bool,
        add_eos: bool,
    ) -> List[int]:
        """
        Encodes a string into a list of token IDs.

        Args:
            s (str): The input string to be encoded.
            add_bos (bool): Whether to prepend the beginning-of-sequence token.
            add_eos (bool): Whether to append the end-of-sequence token.

        Returns:
            list[int]: A list of token IDs.
        """
        assert type(s) is str

        # The tiktoken tokenizer can handle <=400k chars without
        # pyo3_runtime.PanicException.
        TIKTOKEN_MAX_ENCODE_CHARS = 4000_000

        # https://github.com/openai/tiktoken/issues/195
        # Here we iterate over subsequences and split if we exceed the limit
        # of max consecutive non-whitespace or whitespace characters.
        MAX_NO_WHITESPACES_CHARS = 250_000

        substrs = (
            substr
            for i in range(0, len(s), TIKTOKEN_MAX_ENCODE_CHARS)
            for substr in self._split_whitespaces_or_nonwhitespaces(
                s[i : i + TIKTOKEN_MAX_ENCODE_CHARS], MAX_NO_WHITESPACES_CHARS
            )
        )
        t: List[int] = []
        for substr in substrs:
            t.extend(
                self.model.encode(
                    substr,
                    allowed_special="all",
                    disallowed_special=(),
                )
            )
        if add_bos:
            t.insert(0, self.bos_id)
        if add_eos:
            t.append(self.eos_id)
        return t

    def decode(self, t: Sequence[int]) -> str:
        """
        Decodes a list of token IDs into a string.

        Args:
            t (List[int]): The list of token IDs to be decoded.

        Returns:
            str: The decoded string.
        """
        # Typecast is safe here. Tiktoken doesn't do anything list-related with the sequence.
        return self.model.decode(cast(List[int], t))

    @staticmethod
    def _split_whitespaces_or_nonwhitespaces(
        s: str, max_consecutive_slice_len: int
    ) -> Iterator[str]:
        """
        Splits the string `s` so that each substring contains no more than `max_consecutive_slice_len`
        consecutive whitespaces or consecutive non-whitespaces.
        """
        current_slice_len = 0
        current_slice_is_space = s[0].isspace() if len(s) > 0 else False
        slice_start = 0

        for i in range(len(s)):
            is_now_space = s[i].isspace()

            if current_slice_is_space ^ is_now_space:
                current_slice_len = 1
                current_slice_is_space = is_now_space
            else:
                current_slice_len += 1
                if current_slice_len > max_consecutive_slice_len:
                    yield s[slice_start:i]
                    slice_start = i
                    current_slice_len = 1
        yield s[slice_start:]


@dataclass
class PLMTokenizedSample:
    is_valid: bool = True
    text_ids: List[int] = field(default_factory=list)
    image_pos: List[int] = field(default_factory=list)
    response_pos: List[int] = field(default_factory=list)
    num_media_chunks: int = 0


# Note that PLM using LLaMA (3.1 and 3.2) as base LLM.
class PLMTokenizer(Llama3Tokenizer):
    def __init__(
        self,
        model_path: str,
        patch_size: Optional[int] = None,
        pooling_ratio: Optional[float] = None,
        seq_len: Optional[int] = 2048,
        conversation_format: Optional[str] = "plm_sft",
        image_token: Optional[str] = "<|image|>",
        bos_token: Optional[str] = "<|begin_of_text|>",
        eos_token: Optional[str] = "<|end_of_text|>",
    ):
        super().__init__(model_path=model_path)

        self.patch_size = patch_size
        self.pooling_ratio = pooling_ratio
        self.seq_len = seq_len
        self.conversation_template = REGISTERED_CONVS[conversation_format]
        self.image_token = image_token

        self.bos_token_id = self.special_tokens[bos_token]
        self.eos_token_id = self.special_tokens[eos_token]
        self.pad_token_id = self.pad_id
        self.image_token_id = self.special_tokens[self.image_token]
        self.eos_id = self.eos_token_id
        self.n_words = self.n_words

    def __call__(
        self,
        conversations: List[Any],
        media: Optional[torch.Tensor] = None,
        media_type: Optional[str] = "image",
    ) -> PLMTokenizedSample:
        conv_template = self.conversation_template.copy()
        assert self.image_token == conv_template.image_token
        conv_template.add_conv(deepcopy(conversations))

        num_media_chunks = media.size(0)
        if media_type in ["image", "multi_image", "video"]:
            assert self.patch_size is not None
            assert self.pooling_ratio is not None
            width, height = media.size(-2), media.size(-1)
            num_patches = int(
                (width // self.patch_size // self.pooling_ratio)
                * (height // self.patch_size // self.pooling_ratio)
            )
            dialog = conv_template.get_conversation_dict_list(
                num_images=num_media_chunks,
                num_patches=num_patches,
                media_type=media_type,
            )
        elif media_type == "text":
            # This is text-only sample
            dialog = conv_template.get_conversation_dict_list(
                num_images=0,
                num_patches=0,
                media_type=media_type,
            )
        else:
            NotImplementedError(
                f"The supported media types are ['image', 'multi_image', 'video', 'text'], \
                                but found {media_type} which is not supported"
            )

        text_ids = []
        source_ids = []
        response_ids = []
        response_pos = []
        for msg in dialog:
            for role, text in msg.items():
                tokens = self.encode(text, add_bos=False, add_eos=False)
                if role == "assistant":
                    response_ids.extend(tokens)
                else:
                    source_ids.extend(tokens)
                if (
                    len(text_ids) + len(source_ids) + len(response_ids) + 1
                    > self.seq_len
                ):
                    if len(text_ids) == 0:
                        return PLMTokenizedSample(is_valid=False)
                    logger.info(f"Truncated text length to {len(text_ids) + 1}")
                    break
                text_ids.extend(source_ids)
                response_pos.extend(
                    [i + len(text_ids) for i in range(len(response_ids))]
                )
                text_ids.extend(response_ids)
                source_ids = []
                response_ids = []

        image_pos = [i for i, t in enumerate(text_ids) if t == self.image_token_id]
        return PLMTokenizedSample(
            text_ids=text_ids,
            image_pos=image_pos,
            response_pos=response_pos,
            num_media_chunks=num_media_chunks,
        )

    def _tokenize_for_generation(
        self,
        question: List[Any],
        media: Optional[torch.Tensor] = None,
    ):
        if media is not None:
            width, height = media.size(-2), media.size(-1)
            num_patches = int(
                (width // self.patch_size // self.pooling_ratio)
                * (height // self.patch_size // self.pooling_ratio)
            )
            prompt = self.conversation_template.get_generation_prompt(
                question, num_images=len(media), num_patches=num_patches
            )
            text_ids = self.encode(prompt, add_bos=False, add_eos=False)
            image_pos = [i for i, t in enumerate(text_ids) if t == self.image_token_id]
        else:
            raise NotImplementedError(f"Text-only inference is not supported yet.")

        return text_ids, image_pos

    def decode_batch(self, tokens: torch.Tensor) -> List[str]:
        return [self.decode(tokens[i].tolist()) for i in range(tokens.size(0))]


def build_tokenizer(name: str, path: str, **kwargs) -> Tokenizer:
    if name == "llama3":
        return Llama3Tokenizer(path)
    elif name == "plmchat":
        return PLMTokenizer(path, **kwargs)
    else:
        raise NotImplementedError(f"{name} tokenizer type is not implemented")



"""Tests for t2v_compbench module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_t2v_compbench_basics():
    from ayase.modules.t2v_compbench import T2VCompBenchModule
    _test_module_basics(T2VCompBenchModule, "t2v_compbench")

def test_t2v_compbench_video(video_sample):
    from ayase.modules.t2v_compbench import T2VCompBenchModule
    video_sample.quality_metrics = QualityMetrics()
    m = T2VCompBenchModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.

import atexit
import contextlib
import logging
import multiprocessing as mp
import os
import random
import shutil
import signal
import socket
import subprocess
import sys
import tempfile
from dataclasses import asdict, dataclass
from functools import lru_cache, partial, reduce
from itertools import chain
from typing import List, Optional, Tuple, Union

import torch
# For no recompute ops
import xformers.ops
from torch import distributed as dist
from torch.distributed import ReduceOp
from torch.distributed._composable.fsdp import (MixedPrecisionPolicy,
                                                fully_shard)
from torch.distributed._tensor import DTensor
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import \
    checkpoint_wrapper
from torch.distributed.device_mesh import DeviceMesh, init_device_mesh
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.checkpoint import (CheckpointPolicy,
                                    create_selective_checkpoint_contexts)

logger = logging.getLogger()

# for selective AC
default_no_recompute_ops = {
    torch.ops.aten.mm.default,
    torch.ops.aten._scaled_mm.default,
    torch.ops.aten._scaled_dot_product_efficient_attention.default,
    torch.ops.aten._scaled_dot_product_flash_attention.default,
    torch.ops.c10d_functional.reduce_scatter_tensor.default,
    torch.ops.xformers_flash.flash_fwd.default,
}


@dataclass
class DistributedArgs:
    dp_shard: int = (
        1  # In how many shard to split the model weight. Typically number gpu in a node.
    )
    dp_replicate: int = (
        1  # How many times to replicate the model weight. Typically number of nodes.
    )
    tp_size: int = 1
    selective_activation_checkpointing: bool = False
    full_activation_checkpointing: bool = False
    compile: bool = False
    fsdp_type: str = "no_shard"
    model_dtype: str = "bf16"

    matmul_allow_tf32: bool = False
    allow_bf16_reduced_precision_reduction = True
    detect_anomaly: bool = False

    compile_cache_size_limit: int = 8

    spawn_method: str = "forkserver"


@dataclass
class EnvironmentArgs:
    # Use GNU openMP (GOMP) instead of Intel OpenMP [Intel Math Kernel Library (MKL)]
    MKL_SERVICE_FORCE_INTEL: str = "GNU"
    OMP_NUM_THREADS: str = "1"
    MKL_NUM_THREADS: str = "1"
    # faster intra-node collectives, seems to be a cluster specific flag
    ENABLE_INTRA_NODE_COMM: str = "0"
    # avoids OOMs with long context
    TORCH_NCCL_AVOID_RECORD_STREAMS: str = "1"
    # increasing NCCL timeout time before having some NCCL error 22 should give a 16s timeout
    NCCL_IB_TIMEOUT: str = "22"
    NCCL_DEBUG: str = "WARN"
    TORCH_NCCL_ASYNC_ERROR_HANDLING: str = "1"
    PYTORCH_CUDA_ALLOC_CONF: str = "expandable_segments:True"


def get_device_mesh(distributed_args: DistributedArgs):
    tp_size = distributed_args.tp_size
    dp_replicate = distributed_args.dp_replicate
    dp_shard = distributed_args.dp_shard

    assert (
        dp_replicate * dp_shard * tp_size == get_world_size()
    ), f"dp_replicate * dp_shard * tp_size ({dp_replicate} * {dp_shard} * {tp_size}) != world_size ({get_world_size()})"

    dims = []
    names = []
    if dp_replicate >= 1:
        dims.append(dp_replicate)
        names.append("dp_replicate")
    if dp_shard > 1 or distributed_args.fsdp_type == "no_shard":
        dims.append(dp_shard)
        names.append("dp_shard")
    if tp_size > 1:
        dims.append(tp_size)
        names.append("tp")
    dims = tuple(dims)
    names = tuple(names)

    return init_device_mesh("cuda", mesh_shape=dims, mesh_dim_names=names)


def dist_max(x: Union[int, float], mesh: DeviceMesh = None):
    tensor = torch.tensor(x).cuda()
    dist.all_reduce(tensor, op=ReduceOp.MAX, group=mesh.get_group() if mesh else None)
    return tensor


def dist_mean(x: Union[int, float], mesh: DeviceMesh = None):
    tensor = torch.tensor(x).cuda()
    dist.all_reduce(tensor, op=ReduceOp.AVG, group=mesh.get_group() if mesh else None)
    return tensor


def dist_mean_dict(x):
    r = dict()
    for k in x:
        r[k] = dist_mean(x[k])
        r[k] = r[k].item() if (r[k].dim() == 0) else r[k].tolist()
    return r


@lru_cache()
def get_is_torch_run() -> bool:
    return os.environ.get("LOCAL_RANK") is not None


@lru_cache()
def get_is_slurm_job() -> bool:
    return "SLURM_JOB_ID" in os.environ and not get_is_torch_run()


@lru_cache()
def get_global_rank() -> int:
    if get_is_torch_run():
        return int(os.environ["RANK"])
    elif get_is_slurm_job():
        return int(os.environ["SLURM_PROCID"])
    else:
        return 0


@lru_cache()
def get_local_rank() -> int:
    if get_is_torch_run():
        return int(os.environ["LOCAL_RANK"])
    elif get_is_slurm_job():
        return int(os.environ["SLURM_LOCALID"])
    else:
        return 0


@lru_cache()
def get_world_size() -> int:
    if get_is_torch_run():
        return int(os.environ["WORLD_SIZE"])
    elif get_is_slurm_job():
        return int(os.environ["SLURM_NTASKS"])
    else:
        return 1


@lru_cache()
def get_is_master() -> bool:
    return get_global_rank() == 0


@lru_cache()
def get_master_port(job_id: int) -> int:
    if get_is_torch_run():
        return int(os.environ["MASTER_PORT"])
    else:
        MIN_MASTER_PORT, MAX_MASTER_PORT = (20000, 60000)
        rng = random.Random(job_id)
        return rng.randint(MIN_MASTER_PORT, MAX_MASTER_PORT)


@lru_cache()
def get_master_addr() -> str:
    if get_is_torch_run():
        return os.environ["MASTER_ADDR"]
    elif get_is_slurm_job():
        hostnames = subprocess.check_output(
            ["scontrol", "show", "hostnames", os.environ["SLURM_JOB_NODELIST"]]
        )
        return hostnames.split()[0].decode("utf-8")
    else:
        return "127.0.0.1"


def setup_env(env_args):
    env_vars = asdict(env_args)

    # When using Triton, it attempts to locate prebuilt kernels in a cache
    # located at ~/.triton/cache, but when that's backed by NFS this can fail
    # with a "OSError: [Errno 116] Stale file handle" error. If we were to set
    # it to a local directory it would belong to the first user who created it
    # and it would fail for the job of any other successive user assigned to
    # that machine. To avoid all this mess we use a temporary per-process cache.
    triton_cache_dir = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, triton_cache_dir, ignore_errors=True)
    env_vars["TRITON_CACHE_DIR"] = triton_cache_dir

    # We change the tmp dir to /scratch in case it's slurm job
    # This avoids filling up the host's usually limited tmpfs
    # A full tmpfs leads to very slow creation of processes and weird bugs
    if get_is_slurm_job():
        new_tmp = f"/scratch/slurm_tmpdir/{os.environ['SLURM_JOB_ID']}"
        if os.path.exists(new_tmp):
            env_vars["TMP_DIR"] = new_tmp

    for name, value in env_vars.items():
        if os.environ.get(name) != str(value):
            os.environ[name] = str(value)
            logger.warning(f"WARNING: Setting {name} to {value}")


def setup_torch_distributed(dist_args):
    """
    Handle single and multi-GPU / multi-node / SLURM jobs.
    Initialize the following variables:
        - global_rank
        - world_size
    """
    mp.set_start_method(dist_args.spawn_method)
    with mp.Manager():
        pass

    local_rank = get_local_rank()

    os.environ["RANK"] = str(get_global_rank())
    os.environ["WORLD_SIZE"] = str(get_world_size())
    os.environ["MASTER_ADDR"] = get_master_addr()
    os.environ["MASTER_PORT"] = str(
        get_master_port(job_id=int(os.environ.get("SLURM_JOB_ID", -1)))
    )

    if get_is_torch_run():
        logger.info(f"Run launched with torchrun, local rank: {local_rank}")
    elif get_is_slurm_job():
        logger.info(f"Run launched with slurm, local rank: {local_rank}")
    else:
        logger.info("Single GPU job")

    logger.info(f"ENV: {os.environ}")

    # set GPU device
    assert 0 <= local_rank < 8
    if dist_args.matmul_allow_tf32:
        torch.backends.cuda.matmul.allow_tf32 = True
        logger.warning(
            f"WARNING: Setting torch.backends.matmul.allow_tf32 to True. This is faster but less accurate."
        )
    torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = (
        dist_args.allow_bf16_reduced_precision_reduction
    )
    if torch.cuda.device_count() > 1:
        torch.cuda.set_device(local_rank)
    torch.distributed.init_process_group(init_method="env://", backend="nccl")
    torch.autograd.set_detect_anomaly(dist_args.detect_anomaly)


def get_module(module, access_string):
    names = access_string.split(sep=".")
    return reduce(getattr, names, module)


def set_module(module, access_string, value):
    names = access_string.split(sep=".")
    parent = reduce(getattr, names[:-1], module)
    setattr(parent, names[-1], value)


def default_fsdp_grouping_plan(n_layers: int) -> List[Tuple[str, bool]]:
    return [("vision_model", True)] + [
        (f"layers.{i}", i < n_layers - 1) for i in range(n_layers)
    ]


def get_default_policy(no_recompute_ops=None):
    no_recompute_ops = no_recompute_ops or default_no_recompute_ops

    def default_policy(ctx, func, *args, **kwargs):
        return (
            CheckpointPolicy.MUST_SAVE
            if func in no_recompute_ops
            else CheckpointPolicy.PREFER_RECOMPUTE
        )

    return default_policy


@torch.no_grad()
def check_model_value_range(
    model: torch.nn.Module, range: float = 1e3, std: float = 1e3
):
    for name, param in chain(model.named_parameters()):
        if isinstance(param, DTensor):
            param = param.to_local()

        if param.numel() == 0:
            logger.warning(
                f"Model parameter {name} is empty, probably because of FSDP sharding"
            )
            continue

        if torch.isnan(param).any() or torch.isinf(param).any():
            logger.warning(f"Model parameter {name} contains NaN or Inf")

        param_range = param.max() - param.min()
        param_std = param.std()

        if param_range > range:
            logger.warning(
                f"Model parameter {name} has a suspiciously large range ({param_range}): please check initialization and init_weights is defined and called"
            )
        if param_std > std:
            logger.warning(
                f"Model parameter {name} has a suspiciously large standard deviation ({param_std}): please check initialization and init_weights is defined and called"
            )
        if (param == 0).all():
            logger.warning(
                f"Model parameter {name} is all zeros: it might be because of a missing initialization"
            )


def init_signal_handler(callable):
    """
    Handle signals sent by SLURM for time limit / pre-emption.
    """
    signal.signal(signal.SIGUSR2, callable)
    logger.warning("Signal handler installed.")


def requeue_slurm_job():
    prod_id = int(os.environ["SLURM_PROCID"])
    logger.warning("Host: %s - Global rank: %i" % (socket.gethostname(), prod_id))
    if prod_id == 0 and os.environ.get("LAUNCH_WITH", "") != "DORA":
        logger.warning("Requeuing job " + os.environ["SLURM_JOB_ID"])
        os.system("scontrol requeue " + os.environ["SLURM_JOB_ID"])
    else:
        logger.warning("Not the master process, no need to requeue.")
    sys.exit(0)


@contextlib.contextmanager
def clean_env():
    distrib_names = (
        "MASTER_ADDR",
        "MASTER_PORT",
        "RANK",
        "WORLD_SIZE",
        "LOCAL_RANK",
        "LOCAL_WORLD_SIZE",
        "TORCHELASTIC_RUN_ID",
        "DORA_FORCE_DISTRIB",
    )
    cluster_env = {
        x: os.environ.pop(x)
        for x in os.environ
        if x.startswith(
            ("SLURM_", "SLURMD_", "SRUN_", "SBATCH_", "SUBMITIT_", "WANDB_")
        )
        or x in distrib_names
    }
    try:
        yield
    finally:
        os.environ.update(cluster_env)


def parallelize_model(
    model,
    device_mesh,
    model_args,
    distributed_args: DistributedArgs,
    fsdp_grouping_plan: Optional[List[Tuple[str, bool]]] = None,
    tp_parallelize=None,
    no_recompute_ops=None,
):
    if distributed_args.tp_size > 1:
        assert (
            distributed_args.fsdp_type == "full_shard"
        ), "Only full shard is supported for TP parallelism"
        assert tp_parallelize is not None, "TP plan is required for TP parallelism"
        assert (
            distributed_args.compile == False
        ), "Compile is not supported for TP parallelism"

        tp_parallelize(model, device_mesh["tp"], model_args, distributed_args)

    param_dtype = dict(fp32=torch.float32, fp16=torch.float16, bf16=torch.bfloat16)[
        distributed_args.model_dtype
    ]
    if (
        distributed_args.fsdp_type == "full_shard"
        or distributed_args.fsdp_type == "no_shard"
    ):
        if distributed_args.fsdp_type == "no_shard":
            assert (
                distributed_args.dp_shard == 1
            ), "dp_shard must be 1 for no_shard fsdp_type"
            assert (
                device_mesh["dp_shard"].size() == 1
            ), "dp_shard must be 1 for no_shard fsdp_type"

        fsdp_config = dict(
            mp_policy=(
                MixedPrecisionPolicy(
                    param_dtype=param_dtype,
                    reduce_dtype=torch.float32,
                )
            ),
            mesh=(
                device_mesh["dp_replicate", "dp_shard"]
                if distributed_args.dp_shard > 1
                or distributed_args.fsdp_type == "no_shard"
                else device_mesh["dp_replicate"]
            ),
        )

        if fsdp_grouping_plan is None:
            # Assume that the model has list of layers and group around it
            fsdp_grouping_plan = default_fsdp_grouping_plan(len(model.layers))

        for path, reshard_after_forward in fsdp_grouping_plan:
            module = get_module(model, path)
            set_module(
                model,
                path,
                fully_shard(
                    module, **fsdp_config, reshard_after_forward=reshard_after_forward
                ),
            )

        model = fully_shard(model, **fsdp_config, reshard_after_forward=True)
    else:
        raise ValueError(f"Invalid fsdp_type: {distributed_args.fsdp_type}")

    if distributed_args.selective_activation_checkpointing:
        assert (
            not distributed_args.full_activation_checkpointing
        ), "Selective activation checkpointing is incompatible with full activation checkpointing"
        model = checkpoint_wrapper(
            model,
            context_fn=partial(
                create_selective_checkpoint_contexts,
                get_default_policy(no_recompute_ops),
            ),
        )

    if distributed_args.full_activation_checkpointing:
        assert (
            not distributed_args.selective_activation_checkpointing
        ), "Full activation checkpointing is incompatible with selective activation checkpointing"
        logger.debug("FULL ACTIVATION CHECKPOINTING on all transformer blocks")
        # model = checkpoint_wrapper(model, preserve_rng_state=False)  # TODO: Make activation checkpointing work.

        for layer_id, transformer_block in model.layers.named_children():
            transformer_block = checkpoint_wrapper(
                transformer_block, preserve_rng_state=False
            )
            model.layers.register_module(layer_id, transformer_block)

        if hasattr(model, "vision_model"):
            for (
                layer_id,
                resblock,
            ) in model.vision_model.transformer.resblocks.named_children():
                resblock = checkpoint_wrapper(resblock, preserve_rng_state=False)
                model.vision_model.transformer.resblocks.register_module(
                    layer_id, resblock
                )

    if distributed_args.compile:
        torch._dynamo.config.cache_size_limit = (
            distributed_args.compile_cache_size_limit
        )
        model = torch.compile(model)

    return model



"""Tests for compressed_vqa_hdr module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_compressed_vqa_hdr_basics():
    from ayase.modules.compressed_vqa_hdr import CompressedVQAHDRModule
    _test_module_basics(CompressedVQAHDRModule, "compressed_vqa_hdr")

def test_compressed_vqa_hdr_no_reference(image_sample):
    from ayase.modules.compressed_vqa_hdr import CompressedVQAHDRModule
    m = CompressedVQAHDRModule()
    result = m.process(image_sample)
    assert result is image_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from codes in torch.utils.data.distributed
# ------------------------------------------------------------------------

import json
import math
import os
from collections import defaultdict

import torch
import torch.distributed as dist

from fvcore.common.timer import Timer
from lvis import LVIS
from torch.utils.data.sampler import Sampler


def load_dataset_dicts(json_file):
    timer = Timer()
    lvis_api = LVIS(json_file)
    if timer.seconds() > 1:
        print("Loading {} takes {:.2f} seconds.".format(json_file, timer.seconds()))

    img_ids = sorted(lvis_api.imgs.keys())
    imgs = lvis_api.load_imgs(img_ids)
    anns = [lvis_api.img_ann_map[img_id] for img_id in img_ids]

    imgs_anns = list(zip(imgs, anns))
    print(
        "Loaded {} images in the LVIS format from {}".format(len(imgs_anns), json_file)
    )
    dataset_dicts = []

    for img_dict, anno_dict_list in imgs_anns:
        record = {}
        image_id = record["image_id"] = img_dict["id"]
        objs = []
        for anno in anno_dict_list:
            # Check that the image_id in this annotation is the same as
            # the image_id we're looking at.
            # This fails only when the data parsing logic or the annotation file is buggy.
            assert anno["image_id"] == image_id
            obj = {}
            # Convert 1-indexed to 0-indexed
            obj["category_id"] = anno["category_id"] - 1

            objs.append(obj)
        record["annotations"] = objs
        dataset_dicts.append(record)

    return dataset_dicts


def repeat_factors_from_category_frequency(dataset_dicts, repeat_thresh, sqrt=True):
    # 1. For each category c, compute the fraction of images that contain it: f(c)
    category_freq = defaultdict(int)
    for dataset_dict in dataset_dicts:  # For each image (without repeats)
        cat_ids = {ann["category_id"] for ann in dataset_dict["annotations"]}
        for cat_id in cat_ids:
            category_freq[cat_id] += 1
    num_images = len(dataset_dicts)
    for k, v in category_freq.items():
        category_freq[k] = v / num_images

    # 2. For each category c, compute the category-level repeat factor:
    #    r(c) = max(1, sqrt(t / f(c)))
    category_rep = {
        cat_id: max(
            1.0,
            (
                math.sqrt(repeat_thresh / cat_freq)
                if sqrt
                else (repeat_thresh / cat_freq)
            ),
        )
        for cat_id, cat_freq in category_freq.items()
    }
    for cat_id in sorted(category_rep.keys()):
        print(
            f"Cat ID {cat_id}: freq={category_freq[cat_id]:.2f}, rep={category_rep[cat_id]:.2f}"
        )

    # 3. For each image I, compute the image-level repeat factor:
    #    r(I) = max_{c in I} r(c)
    rep_factors = []
    for dataset_dict in dataset_dicts:
        cat_ids = {ann["category_id"] for ann in dataset_dict["annotations"]}
        rep_factor = max({category_rep[cat_id] for cat_id in cat_ids}, default=1.0)
        rep_factors.append(rep_factor)

    return torch.tensor(rep_factors, dtype=torch.float32)


class RepeatFactorTrainingSampler(Sampler):
    def __init__(
        self,
        dataset,
        num_replicas=None,
        rank=None,
        local_rank=None,
        local_size=None,
        shuffle=True,
    ):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle

        json_file = (
            "/checkpoint/onevision/peizesun/public_data/d2_data/lvis/lvis_v1_train.json"
        )
        dataset_dicts = load_dataset_dicts(json_file)
        repeat_factors = repeat_factors_from_category_frequency(
            dataset_dicts, repeat_thresh=0.001
        )
        # Split into whole number (_int_part) and fractional (_frac_part) parts.
        self._int_part = torch.trunc(repeat_factors)
        self._frac_part = repeat_factors - self._int_part

    def _get_epoch_indices(self, generator):
        """
        Create a list of dataset indices (with repeats) to use for one epoch.

        Args:
            generator (torch.Generator): pseudo random number generator used for
                stochastic rounding.

        Returns:
            torch.Tensor: list of dataset indices to use in one epoch. Each index
                is repeated based on its calculated repeat factor.
        """
        # Since repeat factors are fractional, we use stochastic rounding so
        # that the target repeat factor is achieved in expectation over the
        # course of training
        rands = torch.rand(len(self._frac_part), generator=generator)
        rep_factors = self._int_part + (rands < self._frac_part).float()
        # Construct a list of indices in which we repeat images as specified
        indices = []
        for dataset_index, rep_factor in enumerate(rep_factors):
            indices.extend([dataset_index] * int(rep_factor.item()))
        return torch.tensor(indices, dtype=torch.int64)

    def __iter__(self):
        if self.shuffle:
            g = torch.Generator()
            g.manual_seed(self.epoch)
            # Sample indices with repeats determined by stochastic rounding; each
            # "epoch" may have a slightly different size due to the rounding.
            rfs_indices = self._get_epoch_indices(g)
            # deterministically shuffle based on epoch
            randperm = torch.randperm(len(rfs_indices), generator=g)
            indices = rfs_indices[randperm].tolist()
        else:
            g = torch.Generator()
            g.manual_seed(0)
            # Sample indices with repeats determined by stochastic rounding; each
            # "epoch" may have a slightly different size due to the rounding.
            rfs_indices = self._get_epoch_indices(g)
            indices = rfs_indices.tolist()

        # add extra samples to make it evenly divisible
        if self.total_size > len(indices):
            indices += indices[: (self.total_size - len(indices))]
            assert len(indices) == self.total_size
            # subsample
            offset = self.num_samples * self.rank
            indices = indices[offset : offset + self.num_samples]
            assert len(indices) == self.num_samples

            return iter(indices)
        else:
            self.num_samples = int(math.ceil(len(indices) * 1.0 / self.num_replicas))
            self.total_size = self.num_samples * self.num_replicas
            indices += indices[: (self.total_size - len(indices))]
            assert len(indices) == self.total_size
            # subsample
            offset = self.num_samples * self.rank
            indices = indices[offset : offset + self.num_samples]
            assert len(indices) == self.num_samples

            return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch


class DistributedSampler(Sampler):
    """Sampler that restricts data loading to a subset of the dataset.
    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such case, each
    process can pass a DistributedSampler instance as a DataLoader sampler,
    and load a subset of the original dataset that is exclusive to it.
    .. note::
        Dataset is assumed to be of constant size.
    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (optional): Number of processes participating in
            distributed training.
        rank (optional): Rank of the current process within num_replicas.
    """

    def __init__(
        self,
        dataset,
        num_replicas=None,
        rank=None,
        local_rank=None,
        local_size=None,
        shuffle=True,
    ):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = 0
        self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch
            g = torch.Generator()
            g.manual_seed(self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = torch.arange(len(self.dataset)).tolist()

        # add extra samples to make it evenly divisible
        indices += indices[: (self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        offset = self.num_samples * self.rank
        indices = indices[offset : offset + self.num_samples]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch


class NodeDistributedSampler(Sampler):
    """Sampler that restricts data loading to a subset of the dataset.
    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such case, each
    process can pass a DistributedSampler instance as a DataLoader sampler,
    and load a subset of the original dataset that is exclusive to it.
    .. note::
        Dataset is assumed to be of constant size.
    Arguments:
        dataset: Dataset used for sampling.
        num_replicas (optional): Number of processes participating in
            distributed training.
        rank (optional): Rank of the current process within num_replicas.
    """

    def __init__(
        self,
        dataset,
        num_replicas=None,
        rank=None,
        local_rank=None,
        local_size=None,
        shuffle=True,
    ):
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if local_rank is None:
            local_rank = int(os.environ.get("LOCAL_RANK", 0))
        if local_size is None:
            local_size = int(os.environ.get("LOCAL_SIZE", 1))
        self.dataset = dataset
        self.shuffle = shuffle
        self.num_replicas = num_replicas
        self.num_parts = local_size
        self.rank = rank
        self.local_rank = local_rank
        self.epoch = 0
        self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas

        self.total_size_parts = self.num_samples * self.num_replicas // self.num_parts

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch
            g = torch.Generator()
            g.manual_seed(self.epoch)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = torch.arange(len(self.dataset)).tolist()
        indices = [i for i in indices if i % self.num_parts == self.local_rank]

        # add extra samples to make it evenly divisible
        indices += indices[: (self.total_size_parts - len(indices))]
        assert len(indices) == self.total_size_parts

        # subsample
        indices = indices[
            self.rank
            // self.num_parts : self.total_size_parts : self.num_replicas
            // self.num_parts
        ]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch



# Copyright (2024) Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import numpy as np
import os
from typing import List, Dict
from tools.color import Color


class AccuracyMetric:
    def __init__(self, dataset_name, verbose=False) -> None:
        self.dataset_name = dataset_name
        self.results = []
        self.invalid_results = []
        self.dataset = []
        self.verbose = verbose
    
    def add(self, data):
        self.dataset.append(data)
    
    def process(self, dataset: List[Dict]):
        if self.dataset_name in ['MVBench', 'TVBench',]:
            return self._process_group_by_subtask(dataset)
        else:
            return self._process(dataset)

    def _process(self, dataset: List[Dict], subtask=None):
        for data in dataset:
            prompt, response, prediction = data['prompt'], data['response'], data['prediction']
            prediction = prediction.replace('(', '').replace(')', '').strip()
            if len(prediction) <= 0:
                success = False
            else:
                prediction = prediction[0]
                if '0'<=prediction<='5':
                    prediction = chr(int(prediction) + ord('A'))
                success = prediction.isupper() and prediction.isalpha() and len(prediction) == 1
            if success:
                rst = {
                    'success': success,
                    'data': data,
                    'result': {'acc': response == prediction}
                }
                if subtask:
                    rst['subtask'] = subtask
                self.results.append(rst)
            else:
                rst = {
                    'success': success,
                    'data': data,
                    'result': {'acc': response == prediction}
                }
                if subtask:
                    rst['subtask'] = subtask
                self.invalid_results.append(rst)
    
    def _process_group_by_subtask(self, dataset: List[Dict]):
        def _group_by_subtask(dataset):
            subtasks = {}
            for data in dataset:
                if data['dataset'] not in subtasks:
                    subtasks[data['dataset']] = []
                subtasks[data['dataset']].append(data)
            return subtasks
        subtasks = _group_by_subtask(dataset)
        for subtask, subdata in subtasks.items():
            self._process(subdata, subtask)

    def summarize_metric(self):
        if self.dataset_name in ['MVBench', 'TVBench',]:
            return self._summarize_metric_by_subtask()
        else:
            return self._summarize_metric()

    def _summarize_metric(self):
        if self.verbose:
            for result in self.results + self.invalid_results:
                print(f"{Color.red('Success: ' + str(result['success']))}")
                print(Color.blue(json.dumps(result['data'], ensure_ascii=False)))
                print(f"{Color.green('Accuracy: ' + str(result['result']['acc']))}")

        accs = []
        for result in self.results:
            acc = result['result']['acc']
            accs.append(acc)
        avg_acc = np.average(accs)
    
        self.eval_records = [
            f'=====Evaluation Summary=====',
            f'Dataset: {self.dataset_name}\tMetric: Accuracy',
            f'#Successful Results: {len(self.results)}\n#Failed Results: {len(self.invalid_results)}',
            f'Accuracy: {round(avg_acc*100, 1)}',
        ]   
        for info in self.eval_records:
            print(info)
    
    def _summarize_metric_by_subtask(self):
        from prettytable import PrettyTable
        self.table = PrettyTable(['Task','Accuracy','Success','Failed'])
        def _group_by_subtask():
            sub_results = {}
            sub_invalid_results = {}
            for data in self.results:
                if data['subtask'] not in sub_results:
                    sub_results[data['subtask']] = []
                sub_results[data['subtask']].append(data)
            for data in self.invalid_results:
                if data['subtask'] not in sub_invalid_results:
                    sub_invalid_results[data['subtask']] = []
                sub_invalid_results[data['subtask']].append(data)
            return sub_results, sub_invalid_results
        sub_results, sub_invalid_results = _group_by_subtask()
        oa_accs = []
        subtasks = list(sub_results.keys())
        subtasks.sort(key=lambda x:f"{x.split('/')[-1].split(' ')[0][0]}{x.split('/')[-1].split(' ')[1][0]}")
        for subtask in subtasks:
            sub_rsts = sub_results[subtask]
            sub_in_rsts = sub_invalid_results.get(subtask, [])
            accs = []
            for result in sub_rsts:
                acc = result['result']['acc']
                accs.append(acc)
                oa_accs.append(acc)
            avg_acc = np.average(accs)
            task_name = f"{subtask.split('/')[-1].split(' ')[0][0]}{subtask.split('/')[-1].split(' ')[1][0]}"
            self.table.add_row([task_name, round(avg_acc*100, 1), len(sub_rsts), len(sub_in_rsts)])
        self.table.add_row(['OVERALL', round(np.average(oa_accs)*100, 1), len(self.results), len(self.invalid_results)])
        print(f'=====Evaluation Summary=====')
        print(self.table)
    
    def save_results(self, pred_path):
        if os.path.isdir(pred_path):
            output_dir = os.path.join(pred_path, 'eval_records')
        else:
            output_dir = os.path.join(os.path.dirname(pred_path), 'eval_records')
        os.makedirs(output_dir, exist_ok=True)
        fout = open(os.path.join(output_dir, f'{self.dataset_name}_eval_result.txt'), 'w')
        if self.dataset_name in ['MVBench', 'TVBench',]:
            print(self.table, file=fout)
        else:
            for info in self.eval_records:
                fout.write(info+'\n')
        fout.close()



from .evaluate_caption_cider import CIDErMetric
from .evaluate_qa_oe_gpt import GPTMetric
from .evaluate_qa_mc import AccuracyMetric
from .evaluate_dream_gpt import DREAMGPTMetric
from .evaluate_video_mme import VideoMMEAccuracyMetric



"""Tests for HDRMAX module."""

from ayase.models import QualityMetrics

from ..conftest import _test_module_basics


def test_hdrmax_basics():
    from ayase.modules.hdrmax import HDRMAXModule

    _test_module_basics(HDRMAXModule, "hdrmax")


def test_hdrmax_skip_without_reference(video_sample):
    from ayase.modules.hdrmax import HDRMAXModule

    video_sample.quality_metrics = QualityMetrics()
    result = HDRMAXModule().process(video_sample)
    assert result is video_sample
    assert result.quality_metrics.hdrmax_score is None


def test_hdrmax_field_exists():
    qm = QualityMetrics()
    assert hasattr(qm, "hdrmax_score")
    assert qm.hdrmax_score is None


def test_hdrmax_field_group():
    qm = QualityMetrics()
    assert qm._FIELD_GROUPS.get("hdrmax_score") == "hdr"



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
Train and eval functions used in main.py
"""
import math
import os
import sys
from typing import Iterable

import torch
import util.misc as utils
from datasets.coco_eval import CocoEvaluator, convert_to_xywh
from datasets.data_prefetcher import data_prefetcher
from datasets.panoptic_eval import PanopticEvaluator
from models.utils_softnms import batched_soft_nms
from util.misc import NestedTensor


# Make sure this is consistent with datasets/coco.py 
# TODO: make it configurable
SCALE_RANGES_DICT = {
    1728: [[0, 10000], [32, 10000], [32, 10000],],
    1824: [[0, 10000], [0, 10000], [64, 10000], [64, 10000],],
}


def filter_boxes(boxes, min_scale, max_scale):
    """
    boxes: (N, 4) shape
    """
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    keep = (w * h > min_scale * min_scale) & (w * h < max_scale * max_scale)
    return keep


@torch.no_grad()
def evaluate_tta(
    model_no_ema,
    criterion,
    postprocessors,
    data_loader,
    base_ds,
    device,
    output_dir,
    test_hflip_aug,
    tta,
    soft_nms,
    ema=None,
    save_result=False,
    save_result_dir="",
    soft_nms_method="quad",
    nms_thresh=0.7,
    quad_scale=0.5,
    lsj_img_size=1824,
):
    model = model_no_ema if ema is None else ema
    model.eval()
    criterion.eval()

    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter(
        "class_error", utils.SmoothedValue(window_size=1, fmt="{value:.2f}")
    )
    header = "Test:"

    iou_types = tuple(k for k in ("segm", "bbox") if k in postprocessors.keys())
    coco_evaluator = CocoEvaluator(base_ds, iou_types)
    # coco_evaluator.coco_eval[iou_types[0]].params.iouThrs = [0, 0.1, 0.5, 0.75]

    SCALE_RANGES = SCALE_RANGES_DICT[lsj_img_size]
    IMAGE_SIZE = [lsj_img_size for _ in range(len(SCALE_RANGES))]
    
    prediction_list = []
    for samples, targets in metric_logger.log_every(data_loader, 10, header):
        samples = samples.to(device)
        targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

        orig_target_sizes = torch.stack([t["orig_size"] for t in targets], dim=0)
        metric_logger.update(loss=0, class_error=0, loss_bbox=0, loss_ce=0)
        ########################### Begin of inference_one_image ###########################
        if tta:
            assert samples.tensors.shape[0] == 1, "tta only supports batch size 1"
            assert (
                samples.tensors.shape[1] % 3 == 0
            ), "tta requires dimensions of samples.tensors to be divisible by 3"

            all_boxes = []
            all_scores = []
            all_classes = []

            num_scales = samples.tensors.shape[1] // 3
            for scale_ind in range(num_scales):
                first_samples = NestedTensor(
                    samples.tensors[
                        :,
                        scale_ind * 3 : (scale_ind + 1) * 3,
                        : IMAGE_SIZE[scale_ind // 2],
                        : IMAGE_SIZE[scale_ind // 2],
                    ],
                    samples.mask[
                        :,
                        scale_ind,
                        : IMAGE_SIZE[scale_ind // 2],
                        : IMAGE_SIZE[scale_ind // 2],
                    ],
                )

                if scale_ind % 2 == 0:
                    ######## no flip #######
                    outputs = model(first_samples)
                    noaug_results = postprocessors["bbox"](
                        outputs, 
                        orig_target_sizes, 
                        soft_nms=soft_nms,
                        method=soft_nms_method,
                        nms_thresh=nms_thresh,
                        quad_scale=quad_scale,
                    )
                    keep = filter_boxes(
                        noaug_results[0]["boxes"], *SCALE_RANGES[scale_ind // 2]
                    )
                    all_boxes.append(noaug_results[0]["boxes"][keep])
                    all_scores.append(noaug_results[0]["scores"][keep])
                    all_classes.append(noaug_results[0]["labels"][keep])
                else:
                    ######## flipped #######
                    flipped_outputs = model(first_samples)
                    flipped_pred_logits = flipped_outputs["pred_logits"]
                    flipped_pred_boxes = flipped_outputs["pred_boxes"]
                    reflipped_pred_boxes = flipped_pred_boxes[
                        :, :, [0, 1, 2, 3]
                    ] * torch.as_tensor([-1, 1, 1, 1]).to(
                        flipped_pred_boxes.device
                    ) + torch.as_tensor(
                        [1, 0, 0, 0]
                    ).to(
                        flipped_pred_boxes.device
                    )
                    new_outputs = {}
                    new_outputs["pred_logits"] = flipped_pred_logits
                    new_outputs["pred_boxes"] = reflipped_pred_boxes
                    new_results = postprocessors["bbox"](
                        new_outputs, 
                        orig_target_sizes, 
                        soft_nms=soft_nms,
                        method=soft_nms_method,
                        nms_thresh=nms_thresh,
                        quad_scale=quad_scale,
                    )
                    keep = filter_boxes(
                        new_results[0]["boxes"], *SCALE_RANGES[scale_ind // 2]
                    )
                    all_boxes.append(new_results[0]["boxes"][keep])
                    all_scores.append(new_results[0]["scores"][keep])
                    all_classes.append(new_results[0]["labels"][keep])

            ######## merge #######
            all_boxes = torch.cat(all_boxes, dim=0)
            all_scores = torch.cat(all_scores, dim=0)
            all_classes = torch.cat(all_classes, dim=0)

            keep_inds, updated_scores = batched_soft_nms(
                all_boxes,
                all_scores,
                all_classes,
                method=soft_nms_method,
                threshold=nms_thresh,
                quad_scale=quad_scale,
            )
            merged_scores = updated_scores
            merged_classes = all_classes[keep_inds]
            merged_boxes = all_boxes[keep_inds]

            results = [
                {
                    "boxes": merged_boxes,
                    "scores": merged_scores,
                    "labels": merged_classes,
                }
            ]
        else:
            outputs = model(samples)
            results = postprocessors["bbox"](outputs, orig_target_sizes)

        ########################### End of inference_one_image ###########################
        res = {
            target["image_id"].item(): output
            for target, output in zip(targets, results)
        }
        if coco_evaluator is not None:
            coco_evaluator.update(res)

        for target, output in zip(targets, results):
            res_cpu = {
                target["image_id"].item(): {
                    "boxes": output["boxes"].cpu(),
                    "labels": output["labels"].cpu(),
                    "scores": output["scores"].cpu(),
                }
            }
            prediction_list.append(res_cpu)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)

    if save_result:
        from torch import distributed as dist

        os.makedirs(save_result_dir, exist_ok=True)

        rank = dist.get_rank()
        torch.save(
            prediction_list,
            os.path.join(save_result_dir, f"val2017_prediction_{rank}.pth"),
        )

    if coco_evaluator is not None:
        coco_evaluator.synchronize_between_processes()

    # accumulate predictions from all images
    if coco_evaluator is not None:
        coco_evaluator.accumulate()
        coco_evaluator.summarize()

    stats = {k: meter.global_avg for k, meter in metric_logger.meters.items()}
    if coco_evaluator is not None:
        if "bbox" in postprocessors.keys():
            stats["coco_eval_bbox"] = coco_evaluator.coco_eval["bbox"].stats.tolist()
    return stats, coco_evaluator



import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

class ExposureModule(PipelineModule):
    name = "exposure"
    description = "Checks for overexposure, underexposure, and low contrast using histograms"
    default_config = {
        "overexposure_threshold": 0.3,
        "underexposure_threshold": 0.3,
        "contrast_threshold": 30.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        # Configurable thresholds
        self.overexposure_threshold = self.config.get("overexposure_threshold", 0.3) # Max % of pixels > 240
        self.underexposure_threshold = self.config.get("underexposure_threshold", 0.3) # Max % of pixels < 15
        self.contrast_threshold = self.config.get("contrast_threshold", 30.0) # Min std dev of brightness

    def process(self, sample: Sample) -> Sample:
        image = self._load_image(sample)
        if image is None:
            return sample

        try:
            # Convert to Gray/Luma
            gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            total_pixels = gray.size
            
            # 1. Underexposure Check
            # Count pixels < 15
            dark_pixels = np.count_nonzero(gray < 15)
            dark_ratio = dark_pixels / total_pixels
            
            if dark_ratio > self.underexposure_threshold:
                 sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Severe Underexposure: {dark_ratio*100:.1f}% pixels are dark",
                        details={"dark_ratio": float(dark_ratio)},
                        recommendation="Adjust brightness levels or apply gamma correction. Discard if details are lost in shadows."
                    )
                )

            # 2. Overexposure Check
            # Count pixels > 240
            bright_pixels = np.count_nonzero(gray > 240)
            bright_ratio = bright_pixels / total_pixels
            
            if bright_ratio > self.overexposure_threshold:
                 sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Severe Overexposure: {bright_ratio*100:.1f}% pixels are bright (blown out)",
                        details={"bright_ratio": float(bright_ratio)},
                        recommendation="Reduce highlights or apply tone mapping. Discard if highlights are clipped and unrecoverable."
                    )
                )

            # 3. Low Contrast Check
            # Standard deviation of the histogram
            std_dev = np.std(gray)
            if std_dev < self.contrast_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.INFO,
                        message=f"Low Contrast: {std_dev:.1f} (Image might look washed out or flat)",
                        details={"contrast_std": float(std_dev)},
                        recommendation="Increase contrast or use histogram equalization to improve dynamic range."
                    )
                )

        except Exception as e:
            logger.warning(f"Exposure check failed: {e}")

        return sample

    def _load_image(self, sample: Sample) -> Optional[np.ndarray]:
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_count // 2)
                ret, frame = cap.read()
                cap.release()
                return frame if ret else None
            else:
                return cv2.imread(str(sample.path))
        except Exception:
            return None



import json
import os
import sqlite3
import time

import numpy as np


def convert_to_sqlite_db(src_path: str, dst_path: str, media_type: str):
    """TODO: Docstring for convert_to_sqlite_db.

    Args:
        src_path (str): The src json annotation file path.
        dst_path (str): The saved sqlite db path.
        media_type (str): The media type. Either "image" or "video".

    """

    # con = sqlite3.connect("file:"+dst_path+"?mode=ro",uri=True)
    con = sqlite3.connect(dst_path)
    cur = con.cursor()
    print(f"creating table")
    cur.execute("DROP TABLE IF EXISTS annos")
    table_sql = f""" CREATE TABLE IF NOT EXISTS `annos` (
                                        `id` integer PRIMARY KEY,
                                        `{media_type}` text,
                                        `caption` text
                                        )"""
    print(table_sql)
    cur.execute(table_sql)

    with open(src_path, "r") as f:
        anno_list = json.load(f)
    filenames = [anno[media_type] for anno in anno_list]
    captions = [anno["caption"] for anno in anno_list]
    ids = list(range(len(filenames)))
    records = list(zip(ids, filenames, captions))

    cur.executemany(f"INSERT INTO annos (id, {media_type}, caption) VALUES (?,?,?)", records)
    con.commit()
    con.close()


def read_sqlite(db_path):
    """TODO: Docstring for read_sqlite.

    Args:
        db_path (TODO): TODO

    Returns: TODO

    """
    con = sqlite3.connect("file:" + db_path + "?mode=ro", uri=True)
    cur = con.cursor()
    ids = cur.execute("SELECT id FROM annos").fetchall()
    ids = [id[0] for id in ids]
    print("number medias:", len(ids))
    np.random.shuffle(ids)
    for id in ids[:100]:
        t1 = time.time()
        query = f"SELECT * FROM annos WHERE id = {id};"
        res = cur.execute(query)
        newid, filename, caption = res.fetchone()
        t2 = time.time()
        print(f"time: {t2-t1}s", id, newid, filename, caption)
    con.close()


def convert(json_filename, media_type):
    """convert json annotations to sqlite.
    Returns: TODO

    """
    print(f"\n--------converting {filename}----------------")
    # src_path = os.path.join(os.environ["SL_DATA_DIR"], "anno_pretrain", json_filename)
    path = 'your_data_path/anno/anno_pretrain'
    src_path = os.path.join(path, json_filename)
    dst_path = src_path.replace(".json", ".sqlite.db")
    convert_to_sqlite_db(src_path, dst_path, media_type)
    read_sqlite(dst_path)


if __name__ == "__main__":
    filenames = [
        # ["cc12m.json", "image"],
        ["cc3m_train.json", "image"],
        # ["cc3m_val.json", "image"],
        # ["coco.json", "image"],
        # ["sbu.json", "image"],
        # ["vg.json", "image"],
        # ["webvid_10m_train.json", "video"],
        # ["webvid_10m_val.json", "video"],
        ["webvid_train.json", "video"],
    ]
    for filename, media_type in filenames:
        convert(filename, media_type)



# Code from https://github.com/SsnL/dataset-distillation/blob/master/datasets/pascal_voc.py , thanks to the authors
"""Dataset setting and data loader for PASCAL VOC 2007 as a classification task.

Modified from
https://github.com/Cadene/pretrained-models.pytorch/blob/56aa8c921819d14fb36d7248ab71e191b37cb146/pretrainedmodels/datasets/voc.py
"""

import os
import os.path
import tarfile
import xml.etree.ElementTree as ET
from urllib.parse import urlparse

import torch
import torch.utils.data as data
import torchvision
from PIL import Image

object_categories = [
    "aeroplane",
    "bicycle",
    "bird",
    "boat",
    "bottle",
    "bus",
    "car",
    "cat",
    "chair",
    "cow",
    "diningtable",
    "dog",
    "horse",
    "motorbike",
    "person",
    "pottedplant",
    "sheep",
    "sofa",
    "train",
    "tvmonitor",
]

category_to_idx = {c: i for i, c in enumerate(object_categories)}

urls = {
    "devkit": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCdevkit_08-Jun-2007.tar",
    "trainval_2007": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar",
    "test_images_2007": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtest_06-Nov-2007.tar",
    "test_anno_2007": "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtestnoimgs_06-Nov-2007.tar",
}


def download_url(url, path):
    root, filename = os.path.split(path)
    torchvision.datasets.utils.download_url(url, root=root, filename=filename, md5=None)


def download_voc2007(root):
    path_devkit = os.path.join(root, "VOCdevkit")
    path_images = os.path.join(root, "VOCdevkit", "VOC2007", "JPEGImages")
    tmpdir = os.path.join(root, "tmp")

    # create directory
    if not os.path.exists(root):
        os.makedirs(root)

    if not os.path.exists(path_devkit):

        if not os.path.exists(tmpdir):
            os.makedirs(tmpdir)

        parts = urlparse(urls["devkit"])
        filename = os.path.basename(parts.path)
        cached_file = os.path.join(tmpdir, filename)

        if not os.path.exists(cached_file):
            download_url(urls["devkit"], cached_file)

        # extract file
        print(
            "[dataset] Extracting tar file {file} to {path}".format(
                file=cached_file, path=root
            )
        )
        cwd = os.getcwd()
        tar = tarfile.open(cached_file, "r")
        os.chdir(root)
        tar.extractall()
        tar.close()
        os.chdir(cwd)
        print("[dataset] Done!")

    # train/val images/annotations
    if not os.path.exists(path_images):

        # download train/val images/annotations
        parts = urlparse(urls["trainval_2007"])
        filename = os.path.basename(parts.path)
        cached_file = os.path.join(tmpdir, filename)

        if not os.path.exists(cached_file):
            download_url(urls["trainval_2007"], cached_file)

        # extract file
        print(
            "[dataset] Extracting tar file {file} to {path}".format(
                file=cached_file, path=root
            )
        )
        cwd = os.getcwd()
        tar = tarfile.open(cached_file, "r")
        os.chdir(root)
        tar.extractall()
        tar.close()
        os.chdir(cwd)
        print("[dataset] Done!")

    # test annotations
    test_anno = os.path.join(path_devkit, "VOC2007/ImageSets/Main/aeroplane_test.txt")
    if not os.path.exists(test_anno):

        # download test annotations
        parts = urlparse(urls["test_images_2007"])
        filename = os.path.basename(parts.path)
        cached_file = os.path.join(tmpdir, filename)

        if not os.path.exists(cached_file):
            download_url(urls["test_images_2007"], cached_file)

        # extract file
        print(
            "[dataset] Extracting tar file {file} to {path}".format(
                file=cached_file, path=root
            )
        )
        cwd = os.getcwd()
        tar = tarfile.open(cached_file, "r")
        os.chdir(root)
        tar.extractall()
        tar.close()
        os.chdir(cwd)
        print("[dataset] Done!")

    # test images
    test_image = os.path.join(path_devkit, "VOC2007/JPEGImages/000001.jpg")
    if not os.path.exists(test_image):

        # download test images
        parts = urlparse(urls["test_anno_2007"])
        filename = os.path.basename(parts.path)
        cached_file = os.path.join(tmpdir, filename)

        if not os.path.exists(cached_file):
            download_url(urls["test_anno_2007"], cached_file)

        # extract file
        print(
            "[dataset] Extracting tar file {file} to {path}".format(
                file=cached_file, path=root
            )
        )
        cwd = os.getcwd()
        tar = tarfile.open(cached_file, "r")
        os.chdir(root)
        tar.extractall()
        tar.close()
        os.chdir(cwd)
        print("[dataset] Done!")


def read_split(root, dataset, split):
    base_path = os.path.join(root, "VOCdevkit", dataset, "ImageSets", "Main")
    filename = os.path.join(base_path, object_categories[0] + "_" + split + ".txt")

    with open(filename, "r") as f:
        paths = []
        for line in f.readlines():
            line = line.strip().split()
            if len(line) > 0:
                assert len(line) == 2
                paths.append(line[0])

        return tuple(paths)


def read_bndbox(root, dataset, paths):
    xml_base = os.path.join(root, "VOCdevkit", dataset, "Annotations")
    instances = []
    for path in paths:
        xml = ET.parse(os.path.join(xml_base, path + ".xml"))
        for obj in xml.findall("object"):
            c = obj[0]
            assert c.tag == "name", c.tag
            c = category_to_idx[c.text]
            bndbox = obj.find("bndbox")
            xmin = int(bndbox[0].text)  # left
            ymin = int(bndbox[1].text)  # top
            xmax = int(bndbox[2].text)  # right
            ymax = int(bndbox[3].text)  # bottom
            instances.append((path, (xmin, ymin, xmax, ymax), c))
    return instances


class PASCALVoc2007(data.Dataset):
    """
    Multi-label classification problem for voc2007
    labels are of one hot of shape (C,), denoting the presence/absence
    of each class in each image, where C is the number of classes.
    """

    def __init__(
        self, root, set, transform=None, download=False, target_transform=None
    ):
        self.root = root
        self.path_devkit = os.path.join(root, "VOCdevkit")
        self.path_images = os.path.join(root, "VOCdevkit", "VOC2007", "JPEGImages")
        self.transform = transform
        self.target_transform = target_transform

        # download dataset
        if download:
            download_voc2007(self.root)

        paths = read_split(self.root, "VOC2007", set)
        bndboxes = read_bndbox(self.root, "VOC2007", paths)
        labels = torch.zeros(len(paths), len(object_categories))
        path_index = {}
        for i, p in enumerate(paths):
            path_index[p] = i
        for path, bbox, c in bndboxes:
            labels[path_index[path], c] = 1
        self.labels = labels
        self.classes = object_categories
        self.paths = paths

    def __getitem__(self, index):
        path = self.paths[index]
        img = Image.open(os.path.join(self.path_images, path + ".jpg")).convert("RGB")
        target = self.labels[index]
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self):
        return len(self.paths)


class PASCALVoc2007Cropped(data.Dataset):
    """
    voc2007 is originally object detection and multi-label.
    In this version, we just convert it to single-label per image classification
    problem by looping over bounding boxes in the dataset and cropping the relevant
    object.
    """

    def __init__(
        self, root, set, transform=None, download=False, target_transform=None
    ):
        self.root = root
        self.path_devkit = os.path.join(root, "VOCdevkit")
        self.path_images = os.path.join(root, "VOCdevkit", "VOC2007", "JPEGImages")
        self.transform = transform
        self.target_transform = target_transform

        # download dataset
        if download:
            download_voc2007(self.root)

        paths = read_split(self.root, "VOC2007", set)
        self.bndboxes = read_bndbox(self.root, "VOC2007", paths)
        self.classes = object_categories

        print(
            "[dataset] VOC 2007 classification set=%s number of classes=%d  number of bndboxes=%d"
            % (set, len(self.classes), len(self.bndboxes))
        )

    def __getitem__(self, index):
        path, crop, target = self.bndboxes[index]
        img = Image.open(os.path.join(self.path_images, path + ".jpg")).convert("RGB")
        img = img.crop(crop)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def __len__(self):
        return len(self.bndboxes)
